;; forward decls

(define-generic-function c-name)

;; Modular Compiler

,(use paths mlink tables)
,(use objsys)

;,(use compiler)

;;----------------------------------------------------------------
;;
;;  blech!  our REPL module system makes everything shared writable
;;
;;  since we are a compiler, insulate ourselves from redefinitions...
;;

(define load
  (let ((load load))
    (lambda (arg)
      (format #t "loading => ~s\n" arg)
      (load arg))))

(load "../modules/compiler/compiler.scm")

;; this has to come before the "insulation", because otherwise
;; our code will pick up the wrong value of *mifio-defs*

(define-class <definer> (<special-form>))

(define (add-mifio-class name class)
  (set! *mifio-defs* (cons (cons name class) *mifio-defs*))
  (set! *mif-load-defs* #f)
  (set! *mif-save-defs* #f))

(mifio-class "<definer>" <definer>)
(if (not (assoc "<substitution>" *mifio-defs*))
    (mifio-class "<substitution>" <substitution>))

(let ((tbl (table *self*)))
  (for-each (lambda (name tlb)
	      (table-insert! tbl name (clone tlb)))
	    (key-sequence tbl)
	    (value-sequence tbl)))

;;----------------------------------------------------------------

;; collect initialization procedures all together
 
(define $init-thunks '())

(define-syntax (on-startup . body)
  (set! $init-thunks (append $init-thunks
			     (list (lambda () (begin . body))))))

(load "misc.scm")

;; some other stuff

(define-generic-function link-name)

;;----------------------------------------------------------------
;;  bytecode generation

(load "../modules/codegen/genaml.scm")
(load "../modules/codegen/loops.scm")
(load "../modules/codegen/lexaddr.scm")
(load "../modules/codegen/aml2bc.scm")
(load "../modules/codegen/support.scm")

;;----------------------------------------------------------------
;; C-code generation

(load "cgen/assem.scm")
(load "cgen/estsize.scm")
(load "cgen/wrvinsn.scm")

;;----------------------------------------------------------------

;; these should be thread variables if the compiler is to run in a thread

(define *current-module* #f)
(define *build-context* #f)


;; note:  using the bytecode strategy causes an addition
;;        word in the template to appear after the code-properties
;;        but before the literals (it's the bytecode bvec itself)

(load "modules/imported.scm")
(load "config.scm")

(load "configv.scm")
(load "util/resource.scm")

(load "modules/bldctx.scm")
(load "cgen/parts.scm")
(load "modules/mcfload.scm")
(load "modules/document.scm")
(load "modules/use.scm")
(load "modules/findmodl.scm")

(load "cgen/bldfiles.scm")

(load "util/util.scm")
(load "toplevel/builtin.scm")

(load "toplevel/compileclass.scm")

(load "boot/makeinit.scm")
(load "boot/bootable.scm")
(load "util/config-release.scm")

(load "util/genbcode.scm")

(add-special-form-compiler! top-level-compiler->proc)
(add-special-form-compiler! objsys-compiler->proc)

(define (configure-basis-from-modules m-names)
  (let* ((envt (make-top-level-contour))
	 (tbl (table envt)))
    (for-each (lambda (mn)
		(let ((src-tbl (table (top-level-envt (get-ct-module mn)))))
		  (format #t "  *basis* includes: ~s (~d bdgs)\n"
			  mn
			  (table-size src-tbl))
		  (flush-output-port (current-output-port))
		  (table-for-each
		   src-tbl
		   (lambda (h k v)
		     (table-insert! tbl k v)))))
	      m-names)
    (set-compiler-basis! envt)))

(for-each 
 (lambda (thunk) (thunk))
 $init-thunks)

