
;; compile a "define-glue" construct

;; process the literal-special form "(& name)"
;; which evaluates to the <top-level-var> itself

(define (glue-literal expr envt)
  (if (and (pair? expr)
	   (eq? (car expr) '&))
      (if (and (eq? (length expr) 2)
	       (symbol? (cadr expr)))
	  (ensure-tlv (cadr expr) envt)
	  (error/syntax "bad `&' syntax: ~s" expr))
      (let ((ic (compile expr envt envt 'value)))
	(if (ic-const? ic)
	    (value ic)
	    (error/syntax "define-glue literal `~s' not a ct const"
			  expr)))))

(define (compile-tl-define-glue tl-def tl-envt dyn-envt)
  (let ((body (cddr tl-def))
	(name (caadr tl-def))
	(args (cdadr tl-def))
	(literals '())
	(envt '())
	(template? #f))
    
    (if (eq? (car body) ':template)
	(begin
	  (set! template? #t)
	  (set! body (cdr body))))
    
    (if (eq? (car body) 'envt:)
	(begin
	  (if template?
	      (error/syntax 
	       "~a: `envt:' doesn't make sense for template"
	       name))
	  (set! envt
		(map
		 (lambda (l)
		   (glue-literal l tl-envt))
		 (cadr body)))
	  (set! body (cddr body))))

    (if (eq? (car body) 'literals:)
	(begin
	  (set! literals
		(map
		 (lambda (l)
		   (glue-literal l tl-envt))
		 (cadr body)))
	  (set! body (cddr body))))

    (if *tl-report*
	(format #t  "declaring raw glue: ~a ~a\n" 
		(if template? "template" "procedure")
		name))
    (let ((tmpl (glue-template name args literals body)))
      (set-value! (ensure-writable-tlv name tl-envt)
		  (if template? 
		      tmpl 
		      (make <closure>
			    environment: envt
			    template: tmpl)))
      ;; return #f => generate no initialization code
      #f)))

(define (glue-template name args literals body)
  (let ((tmpl (make-gvec* <template> 
			  0 0 '()
			  literals))
  	(bod (map (lambda (m)
		    (if (c-text? m)
			(c-text->string m)
			(if (pair? m)
			    (cons (car m) 
				    (c-text->string (cadr m)))
			    (error/syntax "bad glue body form: ~s" m))))
		    body)))
	(seq-add! (fluid-ref *code-descriptors*)
		    (make <code-descriptor>
			template: tmpl
			code-properties: (list (list 'function-scope
						     name))
			code: (vector args bod)
			strategy: 'literal-c))
	tmpl))

