/*------------------------------------------------------------------*-C-*-*
 * File:    handc/heapi/saveimag.c
 *
 *          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 *          as part of the RScheme project, licensed for free use
 *
 * Version: 1.11
 * Date:    1996.08.11 22:08:04
 * Build:   v0.7-0.9, 97.01.26
 *
 * Purpose: image saver (0.5 style)
 *------------------------------------------------------------------------*/

#ifdef INCL_SCCSID
char *F40RS_dist_heapi_saveimag_sccsid__()
{ return "%Z%%W% 1.11 %G% 22:08:04"; }
#endif /* INCL_SCCSID */

/*
    The new image saver
*/
#include "heapi.h"
#include "imagfile.h"
#include <rscheme/scheme.h>
#include <rscheme/hashmain.h>
#include <rscheme/hashfn.h>
#include <rscheme/linktype.h>
#include "saveq.h"
#include <string.h>
#include <ctype.h>
#include <rscheme/osglue.h>
#include <rscheme/smemory.h>
#include <stdio.h>
#include <stdlib.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

/* #define DEBUG_0 */
#ifdef NDEBUG
#undef NDEBUG
#endif

/*
   NOTE:  Because we use the pob_class field to store
          our unswizzled pointers, we can't call normal
	  runtim.h functions (because they assert their
	  argument types).  This would only be necessary 
	  in a few cases anyway, so they are hand coded
	  using the following macros
*/
#define SYMBOL_STR		SLOT(0)

#define PART_DESCR_PART_TAG	SLOT(1)
#define PART_DESCR_MODULE_NAME	SLOT(0)

#define PAIR_CAR		SLOT(0)
#define PAIR_CDR		SLOT(1)

#define TEMPLATE_CODE_PTR	SLOT(0)
#define TEMPLATE_LINKAGE	SLOT(1)




static SaveQueue load2_queue[NUM_LOAD2_MODES];
static SaveQueue traversal_queue;

static char *output_ptr;
static obj next_name;
static obj refs_vector;
static obj replacement_table;

#define NUM_USEFUL_LOAD2_MODES (5)

static enum load2_mode useful_load2_modes[NUM_USEFUL_LOAD2_MODES] =
    { LOAD2_CLASS_ONLY, 
      LOAD2_GVEC,
      LOAD2_TEMPLATE,
      LOAD2_PARTCONT,
      LOAD2_MIXVEC_2 };

struct unswizzled_ptr;

typedef void writer_fn( const struct unswizzled_ptr *p );

static void write_gvec( const struct unswizzled_ptr *p );
static void write_array_8( const struct unswizzled_ptr *p );
static void write_symbol( const struct unswizzled_ptr *p );
static void write_part_descr( const struct unswizzled_ptr *p );
static void write_part( const struct unswizzled_ptr *p );
static void write_array_32( const struct unswizzled_ptr *p );
static void write_array_64( const struct unswizzled_ptr *p );
static void write_ref( const struct unswizzled_ptr *p );
static void write_template( const struct unswizzled_ptr *p );
static void write_alloc_area( const struct unswizzled_ptr *p );


struct unswizzled_ptr {
    obj 		thing;		/* the pointer itself */
    obj			orig_class;	/* the original class */
    enum load1_mode	mode;		/* output mode */
    writer_fn		*writer;	/* writer function */
    obj			unswiz;		/* the unswizzled name assigned */
};


#define NUM_CLASS_MODES (9)

/* this table maps the image mode field from class objects
   to load1 & load2 modes and the appropriate writer function
*/

static struct { 
    enum load1_mode mode1;
    enum load2_mode mode2;
    writer_fn	    *writer;
} image_modes[NUM_CLASS_MODES] = {
  { /* 0: gvec */ 	LOAD1_ARRAY32,	LOAD2_GVEC,	  write_gvec },
  { /* 1: bvec */	LOAD1_ARRAY8,	LOAD2_CLASS_ONLY, write_array_8 },
  { /* 2: symbol */	LOAD1_SYMBOL,	LOAD2_NOP,	  write_symbol },
  { /* 3: part */	LOAD1_PART,	LOAD2_NOP,	  write_part },
  { /* 4: template */	LOAD1_ARRAY32,	LOAD2_TEMPLATE,	  write_template },
  { /* 5: part-cont */	LOAD1_ARRAY32,	LOAD2_PARTCONT,	  write_gvec },
  { /* 6: longfloat */	LOAD1_ARRAY64,	LOAD2_CLASS_ONLY, write_array_64 },
  { /* 7: uint32*n */	LOAD1_ARRAY32,	LOAD2_CLASS_ONLY, write_array_32 },
  { /* 8: alloc-area */ LOAD1_ARRAY32,  LOAD2_MIXVEC_2,   write_alloc_area } };

static _rs_inline obj use_next_name( obj *without_tag )
{
obj it;

    *without_tag = next_name;
    it = OBJ(VAL(next_name) + POINTER_TAG);
    next_name = OBJ(VAL(next_name) + (1<<PRIMARY_TAG_SIZE));
    return it;
}

static void *hton_obj( void *data )
{
static obj x;

    x = HOST_TO_BIG_ENDIAN_OBJ( *(obj *)data );
    return &x;
}    

/*  Map all objects through the replacement_table  */

static obj replace( obj thing )
{
  if (!EQ(replacement_table,FALSE_OBJ))
    {
      obj k;
      
      k = objecttable_lookup( replacement_table, obj_hash(thing), thing );
      if (!EQ(k,FALSE_OBJ))
	return replace(k);
    }
  return thing;
}

/*  Create a new name for a heap object
    (ie, one that hasn't been unswizzled yet)
    it may be a REF, however
*/

static obj unswizzle_ptr( obj thing )
{
POBHeader *p =(POBHeader *)PTR_TO_HDRPTR(thing);
obj the_unswizzled_ptr;
struct unswizzled_ptr *travq;
enum load1_mode mode1;
enum load2_mode mode2;
int imode;
writer_fn *writer;
obj the_class;

    the_class = p->pob_class;
    if (TEST_PRIMARY_TAG(the_class,1))
    {
	/* a REF that hasn't been unswizzled yet */
	mode1 = LOAD1_REF;
	mode2 = LOAD2_NOP;
	writer = write_ref;
    }
    else if (OBJ_ISA_PTR(the_class))
    {
	obj rplc = replace(the_class);
	/* a normal heap object that hasn't been unswizzled yet */
	
	imode = class_image_mode( rplc );
	if (imode < 0 || imode >= NUM_CLASS_MODES)
	{
	    fprintf( stderr, "Warning: Class %s has illegal image-mode = %d\n",
			symbol_text( class_name( rplc ) ),
			imode );
	    imode = 0;
	}
	mode1 = image_modes[imode].mode1;
	mode2 = image_modes[imode].mode2;
	writer = image_modes[imode].writer;
    }
    else
    {
	/* an error */
	abort();
	mode1 = LOAD1_REF;	/* these are bogus, to elid */
	mode2 = LOAD2_NOP;      /* warning messages */
	writer = write_ref;
    }

    /* insert us in the appropriate LOAD2 queue */

    the_unswizzled_ptr = use_next_name( &p->pob_class );

    travq = (struct unswizzled_ptr *)enqueue_item( &traversal_queue );
    travq->orig_class = the_class;
    travq->mode = mode1;
    travq->thing = thing;
    travq->writer = writer;
    travq->unswiz = the_unswizzled_ptr;

#ifdef DEBUG_0
    printf( "Swizzling: %#x  (class %#x) mode %d / %d ::= %#x\n", 
	    VAL(travq->thing), 
	    VAL(travq->orig_class), 
	    mode1, 
	    mode2, 
	    VAL(the_unswizzled_ptr) );
#endif /* DEBUG_0 */

    if (mode2 != LOAD2_NOP)
    {
	*(obj *)enqueue_item( &load2_queue[mode2] ) = the_unswizzled_ptr;
    }
    
    return the_unswizzled_ptr;
}

/*  Unswizzle any object  */

static _rs_inline obj unswizzle( obj thing )
{
POBHeader *p;

    thing = replace(thing);
    if (OBJ_ISA_PTR(thing))
    {
	p = (POBHeader *)PTR_TO_HDRPTR(thing);

	if (TEST_PRIMARY_TAG(p->pob_class,0))
	{
	    /* already unswizzled */
	    return OBJ( VAL(p->pob_class) + POINTER_TAG );
	}
	return unswizzle_ptr( thing );
    }
    else
	return thing;
}

static void *alloc_output_frame( UINT_32 size_in_bytes, 
				 obj preamble,
				 enum load1_mode mode )
{
UINT_32 *p;

    p = (UINT_32 *)output_ptr;
    output_ptr += (size_in_bytes + 3 + 12) & ~3;
    *p++ = HOST_TO_BIG_ENDIAN_32( size_in_bytes );
    *p++ = HOST_TO_BIG_ENDIAN_32( mode );
    *p++ = VAL(HOST_TO_BIG_ENDIAN_OBJ( unswizzle(preamble) ));
    return p;
}


void write_gvec( const struct unswizzled_ptr *p )
{
obj gvec = p->thing;
obj *d, *s, *limit;
UINT_32 len;

    len = SIZEOF_PTR( gvec );
    s = (obj *)PTR_TO_DATAPTR( gvec );
    limit = (obj *)(((char *)s) + len);
    
    d = (obj *)alloc_output_frame( len, p->orig_class, LOAD1_ARRAY32 );

    while (s < limit)
	*d++ = HOST_TO_BIG_ENDIAN_OBJ( unswizzle( *s++ ) );
}

static void write_mixvec( const struct unswizzled_ptr *p, UINT_32 nslots )
{
obj mixvec = p->thing;
obj *d, *s;
UINT_32 i, len, *s_word, *d_word, *limit_word;

    len = SIZEOF_PTR( mixvec );
    s = (obj *)PTR_TO_DATAPTR( mixvec );

    d = (obj *)alloc_output_frame( len, p->orig_class, LOAD1_ARRAY32 );

    for (i=0; i<nslots; i++)
      {
	*d++ = HOST_TO_BIG_ENDIAN_OBJ( unswizzle( *s++ ) );
      }

    /* write the rest as an array of 32bits */

    limit_word = (UINT_32 *)(((char *)s) + SIZEOF_PTR(mixvec));
    s_word = (UINT_32 *)s;
    d_word = (UINT_32 *)d;

    while (s_word < limit_word)
	*d_word++ = HOST_TO_BIG_ENDIAN_32( *s_word++ );
}

static void write_alloc_area( const struct unswizzled_ptr *p )
{
  write_mixvec( p, 2 );
}


void write_template( const struct unswizzled_ptr *p )
{
obj tmpl = p->thing;
obj link2;
struct function_descr *fn;
obj *d, *s, *limit;
UINT_32 len;
struct function_descr **pfns;
unsigned i;

    link2 = gvec_read(tmpl,TEMPLATE_LINKAGE);
    if (OBJ_ISA_PTR(link2))
    {
#ifdef DEBUG_0
	printf( "template %#x is already unswizzled\n", VAL(tmpl) );
#endif /* DEBUG_0 */
	/* this template is already unswizzled, so just
	   write it out as a gvec */
	write_gvec( p );
    }
    else if (EQ(link2,ZERO))
    {
	fprintf( stderr, "WARNING: Writing stub template %#x\n", VAL(tmpl) );
	write_gvec( p );
    }
    else
    {
#ifdef DEBUG_0
	printf( "template %#x is being unswizzled\n", VAL(tmpl) );
#endif /* DEBUG_0 */

	fn = (struct function_descr *)OBJ_TO_RAW_PTR(link2);

	/* make sure it's not a stub */

	if (fn->in_part->tag >= STUB_PART_TAG)
	  {
	    template_unstub( tmpl );
	    link2 = gvec_read(tmpl,TEMPLATE_LINKAGE);
	    fn = (struct function_descr *)OBJ_TO_RAW_PTR(link2);
	  }

	if (fn->in_part->unswizzled_as == 0)
	{
	obj unswiz, who_cares;
	struct unswizzled_ptr *travq;
	
	    /* this part does not have a <PartDescr> for it yet */
	    
	    unswiz = use_next_name( &who_cares );
	    
	    fn->in_part->unswizzled_as = VAL(unswiz);
	    
	    /* insert us in the LOAD1 queue */
	
	    travq = (struct unswizzled_ptr *)enqueue_item( &traversal_queue );
	    travq->orig_class = ZERO;		/* N/A */
	    travq->mode = (enum load1_mode)0;	/* N/A */
	    travq->thing = RAW_PTR_TO_OBJ(fn->in_part);
	    travq->writer = write_part_descr;
	    travq->unswiz = unswiz;

#ifdef DEBUG_0
	    printf( "\tNew part %#x (%s) is %#x\n", 
	    		(unsigned)fn->in_part,
			fn->in_part->name, 
			fn->in_part->unswizzled_as );
#endif /* DEBUG_0 */
	}
	    
	
	/* find the function's index in the part */

	pfns = fn->in_part->functions;
	for (i=0; pfns[i]; i++)
	{
	    if (pfns[i] == fn)
		break;
	}
#ifdef DEBUG_0
	if (pfns[i])
	    printf( "\t%s is %s.%d\n", fn->name, fn->in_part->name, i );
#endif /* DEBUG_0 */
	if (!pfns[i])
	  {
	    const char *s;

	    fprintf( stderr, "Warning: (function_descr %p) `", fn );
	    for (s = fn->name; *s; s++)
	      if (isprint(*s))
		fputc( *s, stderr );
	      else
		fprintf( stderr, "\\%03o", *(UINT_8 *)s );
	    fprintf( stderr, "'\nis not found in (part %p) `", fn->in_part );
	    for (s = fn->in_part->name; *s; s++)
	      if (isprint(*s))
		fputc( *s, stderr );
	      else
		fprintf( stderr, "\\%03o", *(UINT_8 *)s );
	    fprintf( stderr, "'\n" );
	  }

	len = SIZEOF_PTR( tmpl );
	s = (obj *)PTR_TO_DATAPTR( tmpl );
	limit = (obj *)(((char *)s) + len);
	
	d = (obj *)alloc_output_frame( len, p->orig_class, LOAD1_ARRAY32 );
    
	*d++ = HOST_TO_BIG_ENDIAN_OBJ( int2fx(i) );
	*d++ = HOST_TO_BIG_ENDIAN_OBJ( OBJ(fn->in_part->unswizzled_as) );
	s += 2;
	while (s < limit)
	    *d++ = HOST_TO_BIG_ENDIAN_OBJ( unswizzle( *s++ ) );
    }
}

void write_array_32( const struct unswizzled_ptr *p )
{
obj gvec = p->thing;
UINT_32 *s, *limit, *d, len;

    len = SIZEOF_PTR( gvec );
    s = (UINT_32 *)PTR_TO_DATAPTR( gvec );
    limit = (UINT_32 *)(((char *)s) + len);

    d = (UINT_32 *)alloc_output_frame( len, p->orig_class, LOAD1_ARRAY32 );
    
    while (s < limit)
	*d++ = HOST_TO_BIG_ENDIAN_32( *s++ );
}


void write_array_64( const struct unswizzled_ptr *p )
{
obj gvec = p->thing;
UINT_32 len;
IEEE_64 *s, *limit, *d;

    len = SIZEOF_PTR( gvec );
    s = (IEEE_64 *)PTR_TO_DATAPTR( gvec );
    limit = (IEEE_64 *)(((char *)s) + len);

    d = (IEEE_64 *)alloc_output_frame( len, p->orig_class,  LOAD1_ARRAY64 );
    
    while (s < limit)
	*d++ = HOST_TO_BIG_ENDIAN_IEEE_64( *s++ );
}

/*  Writes a frame of the form
    +---------------+
    |   size        |
    +---------------+
    |    mode       |
    +---------------+
    |   preamble    |
    +---------------+
    |               |
    :     text      :
    |      ....     |
    +---------------+
   which is the format used by LOAD1_REF, LOAD1_SYMBOL,
   LOAD1_PART, and LOAD1_ARRAY8
   (where preamble is:
   	LOAD1_REF	=> string hash value
	LOAD1_SYMBOL	=> string hash value (NOT same as symbol's hash value)
	LOAD1_PART	=> tag of part
	LOAD1_ARRAY8	=> bytevector class)
   Note that the string ('text') had better be ZERO padded to a
   full word length!
*/

static void write_string_frame( enum load1_mode mode, obj preamble, obj text )
{
void *d;
UINT_32 len;

    len = SIZEOF_PTR( text );
    d = alloc_output_frame( len, preamble, mode );
    memcpy( d, PTR_TO_DATAPTR(text), (len + 3) & ~3 );
}


static void write_array_8( const struct unswizzled_ptr *p )
{
    write_string_frame( LOAD1_ARRAY8, p->orig_class, p->thing );
}

static void write_ref( const struct unswizzled_ptr *p )
{
obj ref_info;

    ref_info = gvec_read( refs_vector, VAL(p->orig_class)-1 );
    write_string_frame( LOAD1_REF, 
    			ZERO,
			gvec_read( ref_info, PAIR_CAR ) );
}

static void write_symbol( const struct unswizzled_ptr *p )
{
    write_string_frame( LOAD1_SYMBOL, 
    			ZERO,
			gvec_read( p->thing, SYMBOL_STR ) );
}

/*
  Here, we are writing out a <PartDescr>.  At load time,
  it will be swizzled to point directly to the part_descr
*/

static void write_part( const struct unswizzled_ptr *p )
{
    write_string_frame( LOAD1_PART,
    			gvec_read( p->thing, PART_DESCR_PART_TAG ),
			gvec_read( p->thing, PART_DESCR_MODULE_NAME ) );
}

/* this is kinda wierd, because we are unswizzling
    something that is NOT a real heap object
    (maybe I should make them real heap objects...
    but at some level, there still has to be a
    connection to C) */

static void write_part_descr( const struct unswizzled_ptr *p )
{
struct part_descr *pd = (struct part_descr *)OBJ_TO_RAW_PTR( p->thing );
char *d;
UINT_32 len;

#ifdef DEBUG_0
    printf( "Writing part %#x", (UINT_32)pd );
    fflush(stdout);
    printf( " (%s) as %#x\n", pd->name, pd->unswizzled_as );
#endif /* DEBUG_0 */

    len = strlen( pd->in_module->name );
    d = (char *)alloc_output_frame( len+1, int2fx(pd->tag), LOAD1_PART );
    memcpy( d, pd->in_module->name, len );
    memset( d + len, 0, 4 - (len & 3) );
}


void save_image( FILE *f, obj entry, obj refs, obj option )
{
SaveQueueIterator sqi;
const struct unswizzled_ptr *p;
char *output_buffer;
UINT_32 i, num_refs;
obj *ref_classes;
size_t rc;
struct file_header hdr;
char temp[FILE_HDR_OFFSET];

    refs_vector = refs;
    replacement_table = FALSE_OBJ;

    hdr.magic = 0;
    hdr.version = 0;
    hdr.num_objects = 0;
    hdr.data_area_length = 0;
    hdr.entry_object = ZERO;
    for (i=0; i<NUM_LOAD2_MODES; i++)
    {
	hdr.load2_offset[i] = 0;
	hdr.load2_count[i] = 0;
    }

    memset( temp, ' ', FILE_HDR_OFFSET );
    if (STRING_P(option))
      {
	int len = string_length(option);

	memcpy( temp, 
	        string_text(option),
	        (len < FILE_HDR_OFFSET) ? len : FILE_HDR_OFFSET );
      }
    else if (OBJ_ISA_PTR(option))
      {
	/*  in this, it should be a hashtable (<object-table>) denoting
	    a remapping of objects in the source data structure
	*/
	replacement_table = option;
      }
    else if (truish(option))
      {
	scheme_error( "save-image: option ~s not understood", 1, option );
      }
      
    rc = fseek( f, 0, SEEK_SET );
    rc = fwrite( temp, 1, FILE_HDR_OFFSET, f );
    rc = fwrite( &hdr, sizeof(struct file_header), 1, f );
    
    assert( ftell(f) == FILE_DATA_OFFSET );

    {
    POBHeader *h;
    obj ref_info;
    
	num_refs = SIZEOF_PTR(refs) / sizeof(obj);
	ref_classes = (obj *)malloc( sizeof(obj) * num_refs );
	for (i=0; i<num_refs; i++)
	{
	    ref_info = gvec_read( refs, SLOT(i) );
	    h = PTR_TO_HDRPTR( gvec_read( ref_info, PAIR_CDR ) );
#ifdef DEBUG_0
	    printf( "Lookaside: REF %#x is class %#x\n", 
	    		VAL(gvec_read( ref_info, PAIR_CDR )), 
			VAL(h->pob_class) );
#endif /* DEBUG_0 */
	    assert( OBJ_ISA_PTR( h->pob_class ) );
	    ref_classes[i] = h->pob_class;
	    h->pob_class = OBJ( (i << PRIMARY_TAG_SIZE) + 1 );
	}
    }

    init_queue( &traversal_queue, 
	        sizeof(struct unswizzled_ptr), 
	        NULL );
    for (i=0; i<NUM_USEFUL_LOAD2_MODES; i++)
    {
    enum load2_mode mode2 = useful_load2_modes[i];
    
	init_queue( &load2_queue[mode2], sizeof(obj), hton_obj );
    }

    output_buffer = (char *)malloc( 10*1024*1024 );
    assert( output_buffer );
    output_ptr = output_buffer;
    next_name = ZERO;

    init_queue_state( &sqi, &traversal_queue );
    hdr.entry_object = HOST_TO_BIG_ENDIAN_OBJ( unswizzle( entry ) );
    
    while ((p = (struct unswizzled_ptr *)dequeue_item( &sqi )) != NULL)
    {
#ifdef DEBUG_0
	printf( "%#x ::= Mode %d : %#x (class %#x)\n",
		VAL(p->unswiz), p->mode, VAL(p->thing), VAL(p->orig_class) );
#endif /* DEBUG_0 */
	p->writer( p );
    }

    /* write out the main data */
    
    fwrite( output_buffer, 1, output_ptr - output_buffer, f );
    hdr.num_objects = HOST_TO_BIG_ENDIAN_32( FIXNUM_TO_RAWINT( next_name ) );
    hdr.magic = HOST_TO_BIG_ENDIAN_32( IMAGE_MAGIC_NUMBER );
    hdr.version = HOST_TO_BIG_ENDIAN_32( IMAGE_VERSION_NUMBER );
    hdr.data_area_length = HOST_TO_BIG_ENDIAN_32( output_ptr - output_buffer );
    
    free( output_buffer );

    /* write out the LOAD2 queues */
    
    for (i=0; i<NUM_USEFUL_LOAD2_MODES; i++)
    {
    UINT_32 n;
    enum load2_mode mode2 = useful_load2_modes[i];
    
	hdr.load2_offset[mode2] = HOST_TO_BIG_ENDIAN_32( ftell( f ) );
	n = fwrite_queue_contents( &load2_queue[mode2], f );
	hdr.load2_count[mode2] = HOST_TO_BIG_ENDIAN_32( n );
#ifdef DEBUG_0
	printf( "Mode %d queue: %u count at %ld\n", 
		    mode2, 
		    n, 
		    hdr.load2_offset[mode2] );
#endif /* DEBUG_0 */
	free_queue( &load2_queue[mode2] );
    }
    
    /* clean up the swizzled objects */
    
    init_queue_state( &sqi, &traversal_queue );
    while ((p = (struct unswizzled_ptr *)dequeue_item(&sqi)) != NULL)
    {
    POBHeader *h;
	/* clean up things that were written as PartDescr's
	   specially, because they don't ever exist as heap objects
	   (but their unswizzled_as flags need to be reset!)
	   (NOTE:  There is obviously something wrong when I 
	    don't have first-class objects for parts (that have
	    object identity)) */
	if (p->writer != write_part_descr)
	{
#ifdef DEBUG_0
	    printf( "Cleaning up: %#x  (class %#x)\n",
			    VAL(p->thing), VAL(p->orig_class) );
#endif /* DEBUG_0 */
	    h = PTR_TO_HDRPTR( p->thing );
	    h->pob_class = p->orig_class;
	}
	else
	{
	struct part_descr *pd = (struct part_descr *)VAL(p->thing);
	
#ifdef DEBUG_0
	    printf( "Cleaning up: %#x  (part %s)\n",
			    VAL(p->thing), pd->name );
#endif /* DEBUG_0 */
	    pd->unswizzled_as = 0;
	}
    }
    
    /* free the memory used by the queues */
    
    free_queue( &traversal_queue );
    
    /* clean up the REFERENCE objects */
    
    {
    POBHeader *h;
    obj ref_info;
    
	for (i=0; i<num_refs; i++)
	{
	    ref_info = gvec_read( refs, SLOT(i) );
	    h = PTR_TO_HDRPTR( gvec_read( ref_info, PAIR_CDR ) );
	    h->pob_class = ref_classes[i];
#ifdef DEBUG_0
	    printf( "Restoring: REF %#x to class %#x\n", 
	    		VAL(gvec_read( ref_info, PAIR_CDR )), 
			VAL(h->pob_class) );
#endif /* DEBUG_0 */
	}
    }
    fseek( f, FILE_HDR_OFFSET, SEEK_SET );
    fwrite( &hdr, sizeof(struct file_header), 1, f );
}

void save_image_file( const char *path, obj entry, obj refs, obj opt )
{
FILE *f = os_fopen( path, "wb" );

    if (!f)
    {
	perror( path );
	return;
    }
    save_image( f, entry, refs, opt );
    fclose( f );
}

