/*-----------------------------------------------------------------*-C-*---
 * File:    handc/runtime/basemath.c
 *
 *          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 *          as part of the RScheme project, licensed for free use
 *
 * Version: 1.3
 * Date:    1996.09.12 11:07:27
 * Build:   v0.7-0.9, 97.01.26
 *
 * Purpose: Arithmetic for basic number system
 *------------------------------------------------------------------------*/

#include <rscheme/scheme.h>

obj basic_plus( obj x, obj y )
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    return FX_ADD(x,y);
	else if (LONGFLOAT_P(y))
	    return make_float( fx2int(x) + extract_float(y) );
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    return make_float( extract_float(x) + fx2int(y) );
	else if (LONGFLOAT_P(y))
	    return make_float( extract_float(x) + extract_float(y) );
    }

    scheme_error( "(base+ ~s ~s): argument not basic", 2, x, y );
    return ZERO;
}

obj basic_minus( obj x, obj y )
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    return FX_SUB(x,y);
	else if (LONGFLOAT_P(y))
	    return make_float( fx2int(x) - extract_float(y) );
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    return make_float( extract_float(x) - fx2int(y) );
	else if (LONGFLOAT_P(y))
	    return make_float( extract_float(x) - extract_float(y) );
    }

    scheme_error( "(base- ~s ~s): argument not basic", 2, x, y );
    return ZERO;
}

obj basic_mul( obj x, obj y )
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    return FX_MUL(x,y);
	else if (LONGFLOAT_P(y))
	    return make_float( fx2int(x) * extract_float(y) );
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    return make_float( extract_float(x) * fx2int(y) );
	else if (LONGFLOAT_P(y))
	    return make_float( extract_float(x) * extract_float(y) );
    }
    scheme_error( "(base* ~s ~s): argument not basic", 2, x, y );
    return ZERO;
}

obj basic_div( obj x, obj y )
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    return make_float( (double)fx2int(x) / (double)fx2int(y) );
	else if (LONGFLOAT_P(y))
	    return make_float( fx2int(x) / extract_float(y) );
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    return make_float( extract_float(x) / fx2int(y) );
	else if (LONGFLOAT_P(y))
	    return make_float( extract_float(x) / extract_float(y) );
    }

    scheme_error( "(base/ ~s ~s): argument not basic", 2, x, y );
    return ZERO;
}

static int float_cmp( IEEE_64 a, IEEE_64 b )
{
  if (a < b)
    return -1;
  else if (a > b)
    return 1;
  else
    return 0;
}

int basic_cmp( obj x, obj y )
{
  if (OBJ_ISA_FIXNUM(x))
    {
      if (OBJ_ISA_FIXNUM(y))
	{
	  if (EQ(x,y))
	    return 0;
	  else if (FX_LT(x,y))
	    return -1;
	  else
	    return 1;
	}
      else if (LONGFLOAT_P(y))
	{
	  return float_cmp( (IEEE_64)fx2int(x), extract_float(y) );
	}
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	  return float_cmp( extract_float(x), (IEEE_64)fx2int(y) );
	else if (LONGFLOAT_P(y))
	  return float_cmp( extract_float(x), extract_float(y) );
    }

    scheme_error( "(base-cmp ~s ~s): argument not basic", 2, x, y );
    return NO;
}
