#|------------------------------------------------------------*-Scheme-*--|
 | File:    bytcodes/aml2bc.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.6
 | Date:    1996.07.25 17:25:07
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: compile abstract machine language into bytecodes
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.6 %G% 17:25:07")

;; analagous to assem.scm...

(define (aml-expr-is-obj? aml-expr)
  (prim-subtype? (lookup-prim-type (car aml-expr))
		 (lookup-prim-type '<obj>)))

(define (aml-expr-is-raw-bool? aml-expr)
  (prim-subtype? (lookup-prim-type (car aml-expr))
		 (lookup-prim-type '<raw-bool>)))

(define (compile-aml-stmt insn)
  (if *debug-byte-codes*
      (format #t "aml-stmt: ~s\n" insn))
    (case (car insn)
	((apply) (aml-apply insn))
	((set!) (aml-set! insn))
	((save) (aml-save insn))
	((restore) (aml-restore insn))
	((label) (def-label (cadr insn)))
	((return) (aml-return insn))
	((bind) (aml-bind insn))
	((unbind) (emit-unbind))
	((jump) (emit-jump (cadr insn) (caddr insn)))
	((bjump) (emit-bjump (cadr insn) (caddr insn)))
	((if) (aml-if insn))
	((seq) (for-each compile-aml-stmt (cdr insn)))
	((do) (compile-aml-expr (cadr insn))
	      (emit-pop))
	((check=) (emit-check= (cadr insn)))
	((check>=) (emit-check>= (cadr insn)))
	((collect>) (emit-collect> (cadr insn)))
	((set-false<) (emit-set-false< (cadr insn)))
	((use-empty-envt) (emit-use-empty-envt))
	((use-function-envt) (emit-use-function-envt))
	;;
	;; a primop statement is generated from a primcall
	;; that returns no values
	;;
	((primop)
	 (let ((primop (cadr insn))
	       (args (cddr insn)))
	   (for-each compile-aml-expr args)
	   (emit-primop primop (length args))))
	(else
	    (abort 'compile-aml-stmt
	           "Unrecognized AML stmt: ~s" insn))))

(define alloc-temp-label
  (let ((i 10000))
    (lambda ()
      (set! i (+ i 1))
      i)))

;; note that if's are at the end of a monotone too,
;; so we don't need to have our own JOIN label

(define (aml-if insn)
  (let ((l (alloc-temp-label)))
    (assert (aml-expr-is-raw-bool? (cadr insn)))
    (compile-aml-expr (cadr insn))
    (emit-branch-if-false l)
    (compile-aml-stmt (caddr insn))
    (def-label l)
    (compile-aml-stmt (cadddr insn))))

(define (reg-ref? aml-expr)
  (let ((aml-expr (cdr aml-expr)))
    (and (eq? (car aml-expr) 'ref)
	 (eq? (caadr aml-expr) 'reg))))

(define (aml-bind insn)
  (assert (every? aml-expr-is-obj? (cdr insn)))
  ;; special case  (bind (ref (reg 0)) (ref (reg 1)) ... (ref (reg N-1)))
  (if (and (every? reg-ref? (cdr insn))
	   (equal? (map cadr (map caddr (cdr insn))) 
		   (range (length (cdr insn)))))
      ;;
      ;; it's a bind of registers from 0 to N-1
      ;;
      (emit-bind-first-regs (length (cdr insn)))
      ;; it's a bind not all of registers, or not the particular
      ;; registers in the right order
       (begin
	 (for-each compile-aml-expr (cdr insn))
	 (emit-bind (length (cdr insn))))))

(define (aml-save insn)
    (emit-save (cadr insn) (caddr insn)))
    
(define (aml-restore insn)
    (emit-restore (cadr insn)))

(define (aml-return insn)
    (emit-return (cadr insn)))
    
(define (aml-apply insn)
    (compile-aml-expr (caddr insn))
    (emit-apply (cadr insn)))

(define (aml-set! insn)
  (assert (aml-expr-is-obj? (caddr insn)))
  (let ((ea (cadr insn))
	(rhs (caddr insn)))
    ;(format #t "aml-set: ea = ~s, rhs = ~s\n" ea rhs)
    ;;
    ;; do a peep-hole optimization
    ;;  throw away things of the form: (set! (reg ,x) (ref (reg ,x)))
    ;;
    (if (and (eq? (car ea) 'reg)
	     (eq? (cadr rhs) 'ref)
	     (eq? (caaddr rhs) 'reg))
	(let ((to (cadr ea))
	      (from (cadadr (cdr rhs))))
	  (if (eq? from to)
	      #t ;; instant NOP
	      (emit-reg-xfer from to)))
	;;
	(begin
	  ;; get the RHS onto the eval stack
	  (compile-aml-expr rhs)
	  (case (car ea)
	    ((reg)  (emit-reg-set (cadr ea)))
	    ((lex-var) (emit-lex-set (cadr ea) (caddr ea)))
	    ((tl-var) (emit-tl-set (cadr ea)))
	    (else   (error/internal "Bad EA in set!: ~s" ea)))))))


(define (compile-aml-expr expr)
  (if *debug-byte-codes*
      (format #t "compile-aml-expr: ~s\n" expr))
    ;; the car of an expr is it's type
    (set! expr (cdr expr))
    (if (pair? expr)
	(case (car expr)
	    ((literal)
		(emit-literal (cadr expr)))
	    ((closure)
	     (emit-closure (cadr expr)))
	    ((this-function)
	     (emit-this-function))
	    ((seq)
		(assert (pair? (cdr expr)))
		(let loop ((i (cdr expr)))
		    (if (null? (cdr i))
			(compile-aml-expr (car i))
			(begin
			    (compile-aml-expr (car i))
			    (emit-pop)))))
	    ((primop)
	    	(let ((primop (cadr expr))
		      (args (cddr expr)))
		  ;; special case the application of the
		  ;; raw-int->fixnum primop to a raw-int constant
		  (if (and (eq? (name primop) 'raw-int->fixnum)
			   (eq? (length args) 1)
			   (eq? (cadar args) 'int))
		      ;;
		      ;; it's a (primop raw-int->fixnum (<raw-int> int N))
		      ;;
		      (emit-fixnum (caddar args))
		      ;;
		      ;; it's some other primop...
		      ;;
		      (begin
			(for-each compile-aml-expr args)
			(emit-primop primop (length args))))))
	    ((ref)
		(let ((ea (cadr expr)))
		    (case (car ea)
			((reg) (emit-reg-ref (cadr ea)))
			((lex-var) (emit-lex-ref (cadr ea) (caddr ea)))
			((tl-var) (emit-tl-ref (cadr ea)))

;;
;; user-defined root variables not yet/no longer supported in 0.6
;;
;;			((root) (emit-ref-root (cadr ea)))

			(else (error/internal "Bad EA in ref: ~s" ea)))))
	    ((immob)
	    	(emit-immob (cadr expr)))
	    ((int)
	        (emit-raw-int (cadr expr)))
	    ((raw-bool)
	        (emit-raw-bool (cadr expr)))
	    (else
	     (error/internal "Bad EXPR: ~s" expr)))
	; plain immobs no longer supported
	;(emit-immob expr)
	(error/internal "Bad EXPR: ~s" expr)))
