#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/compiler/kwdlist.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.3
 | Date:    1996.04.18 20:27:40
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: compile keyword lists
 |------------------------------------------------------------------------|
 | Notes:
 |     *** This file is included by the offline compiler ***
 `------------------------------------------------------------------------|#





;; returns a list of <icode> for		foo: bar baz: blech
;; that is the same as would result from
;; 						'foo: bar 'baz: blech
;;
;; (in this version, subst expansions done, because
;;  we might as well -- compile-make is called as from a macro)
;;
;; this function is currently only used for parsing the
;; arguments to `make' after the class expr
;;

(define (compile-keyword-list kw-list lex-envt dyn-envt)
  (cond
   ((null? kw-list)
    '())
   ((and (pair? kw-list)
	 (or (keyword? (car kw-list))
	     (unique-obj? (car kw-list)))
	 (pair? (cdr kw-list)))
    (cons (make <ic-const> 
		value: (car kw-list))
	  (cons (compile (cadr kw-list) lex-envt dyn-envt 'value)
		(compile-keyword-list (cddr kw-list)
				      lex-envt
				      dyn-envt))))
   ((symbol? kw-list)
    ;; it is something like (foo: bar baz: quux . more)
    (let ((bdg (lookup-aliased kw-list lex-envt dyn-envt)))
      (if (substitution? bdg)
	  (compile-keyword-list (expr bdg) (envt bdg) dyn-envt)
	  (error/syntax "at `~s', not a substitution"
			kw-list))))
   (else					      
    (error/syntax "at `~s', keyword/value list ill formed"
		  kw-list))))

;;
;;  for a list of the form
;;    (keyword: value keyword: value ...)
;;  calls the given procedure (proc) for each kwd/val pair

(define (for-each-keyword proc kv-list)
  (let loop ((i kv-list))
    (if (pair? i)
	(if (keyword? (car i))
	    (if (pair? (cdr i))
		(if (proc (car i) (cadr i))
		    (loop (cddr i))
		    (error/syntax "keyword not recognized: ~s" (car i)))
		(error/syntax "missing value after: ~s" (car i)))
	    (error/syntax "not a keyword: ~s" (car i)))
	(if (not (null? i))
	    (error/syntax "keyword list malformed")))))
