#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/corelib/classes.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.17
 | Date:    1996.12.31 16:27:57
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: standard, core system classes
 `------------------------------------------------------------------------|#

;;(file-version-string "%Z%%W% 1.17 %G% 16:27:57")

(define-class <object> () :abstract)

(define-generic-function initialize)
(define-generic-function finalize)   ;; only invoked on registered objects

(define-generic-function display-object)
(define-generic-function write-object)

;;
;;  collection class structured based on Dylan(TM)
;;

(define-class <collection> (<object>) :abstract)
(define-class <sequence> (<collection>) :abstract)

(define-class <vector> (<sequence>))
(define-class <list> (<sequence>) :abstract)
(define-class <pair> (<list>))
(define-class <empty-list> (<list>) :immob)

(define-class <string> (<object>) :bvec)
(define-class <char> (<object>) :abstract)

(define-class <ascii-char> (<char>) :immob)

(define-class <unicode-char> (<char>) :immob)
(define-class <unicode-string> (<object>) :bvec)

(define-class <number> (<object>) :abstract)

(define-class <complex> (<number>) :abstract)

(define-class <real> (<complex>) :abstract)
(define-class <float> (<real>) :abstract)
(define-class <rational> (<real>) :abstract)
(define-class <integer> (<rational>) :abstract)

(define-class <fixnum> (<integer>) :immob)
(define-class <double-float> (<float>) :bvec)
(define-class <long-int> (<integer>) :bvec)

(define-class <boolean> (<object>) :immob)
(define-class <unique-obj> (<object>) :immob)

;;  spare (unused) secondary tag indicators...

(define-class <spare-1> (<object>) :immob)
(define-class <spare-2> (<object>) :immob)
(define-class <spare-3> (<object>) :immob)

(define-class <byte-vector> (<object>) :bvec)

;;
;;  <byte-coded> things are bvecs that contain byte-coded programs
;;  typically, they occur in slot[3] of a <template> 

(define-class <byte-coded> (<object>) :bvec)

;; <symbol>'s are defined here to have image-mode: 0
;; so that when THIS image is being written out
;; (ie, as a boot image), symbols are written out
;; as regular objects instead of in SYMBOL mode

(define-class <symbol> (<object>) #|image-mode: 2|#
  (string-value allocation: internal)
  (hash-value allocation: internal))

;;  instances of <binding-envt> grow to fit the needs...

(define-class <binding-envt> (<object>)
  enclosing)

;; a function is the class of callable objects.
;; it is abstract

(define-class <function> (<object>) :abstract
  environment
  template)

;; a closure is what you get from `lambda'
;; it's a minimal-functionality (sorry) function

(define-class <closure> (<function>))

;; a method is what you get with define-method
;; and is something that's designed to be put in
;; a generic function
;; (in particular, the compilation of a method
;;  may assume that the function specializers have
;;  already been checked (currently, it should only
;;  assume that about the first, because we only
;;  have single-method dispatch))

(define-class <method> (<function>)
  function-specializers
  (sync-method init-value: #f))

;; the environment of a <gf> contains
;; a binding (in LEXADDR(0:0)) which holds a pointer
;; to the <gf> instance itself.  This is because
;; the calling convention doesn't keep track of the
;; closure that got called, so we store a back 
;; pointer in the envt, which DOES get passed (envt_reg)
;;
;; the envt may also hold the cache, since that's
;; faster to access.  otherwise, useful info is hung
;; off the <gf> itself, because they are not created
;; on the fly like <closure>'s are, and it makes more
;; sense to put it there (since they aren't created on
;; the fly, it doesn't hurt to make them big)

(define-class <generic-function> (<function>)
  generic-function-methods
  function-specializers
  generic-function-name)

;; this much information is known to the runtime system...

(define-class <<class>> (<object>)
  class-name
  heap-type
  image-mode
  superclasses)

;; the rest is high-level...

(define-class <<standard-class>> (<<class>>)
  direct-slots
  all-slots
  instance-size
  corresponding-primtype
  class-precedence-list
  (spare-0 init-keyword: #f getter: #f setter: #f init-value: #f)
  (spare-1 init-keyword: #f getter: #f setter: #f init-value: #f)
  (spare-2 init-keyword: #f getter: #f setter: #f init-value: #f)
  (spare-3 init-keyword: #f getter: #f setter: #f init-value: #f)
  (spare-4 init-keyword: #f getter: #f setter: #f init-value: #f))

(define-class <slot-descriptor> (<object>)
    name
    initialization-mode		;; required, optional, prohibited
    init-value
    type-restriction
    (index type: <fixnum>)
    init-keyword
    getter
    setter
    (properties type: <list> init-value: '()))


(define-class <template> (<object>) image-mode: 4
  code-pointer
  linkage-info
  function-descr)

(define-class <partial-continuation> (<object>) image-mode: 5)

(define-class <binding> (<object>)
  name)

(define-class <top-level-var> (<binding>)
  value
  (write-prot init-value: #f
	      type: <boolean>))

(define-class <substitution> (<binding>)
    expr
    envt)   ;; this is the envt. that the expr COMES FROM

(define-class <macro> (<binding>)
    envt
    forms
    else-bdg
    (setter-forms init-value: '()))

(define-class <rewriter> (<binding>)
  rewriter-body    ;; something like '((cons 'foo (cdr form)))
  rewriter-args)   ;; type '(form)

;; a rewriter that has access to it's environment of definition

(define-class <local-rewriter> (<rewriter>)
  rewriter-lex-envt
  rewriter-dyn-envt)

;; note-- this has to be in sync with [base]/compiler/cmplr/bdgs.bbi's <primop>

(define-class <primop> (<binding>)
    (arg-types init-value: '() type: <list>)
    (result-type init-value: '<obj>)
    (rest-type init-value: #f)
    (translations init-value: '() type: <list>)
    (full-procedure-bdg init-value: #f)
    (primop-has-side-effect? init-value: #t)) ;; #f for side-effecting primops



;; a <scope-record> is the compile-time representation
;; of some kind of binding envt.

(define-class <scope-record> (<object>))

(define-class <lexical-contour> (<scope-record>)
  name->bindings
  bindings
  lexical-enclosing
  dynamic-enclosing)

(define-class <top-level-contour> (<scope-record>)
  (table init-value: #f)
  (owner init-value: #f)
  (backing init-value: #f)
  (dirty? init-value: #f))

(define-class <macro-form> (<object>)
    args
    body)

#|
	     (<top-level-envt> :gvec)
	     (<imported-binding> :gvec)

	     (<patch> :gvec)
	     (<part-descr> :gvec image-mode: 3)

	     (<module> :gvec)
	     (<binding> :abstract
			(<primop> :gvec)
			(<top-level-var> :gvec)
			(<definer> :gvec))
	     (<byte-coded> :bvec)

	     (<link-cmd> :abstract
			 (<link-xform> :gvec)
			 (<link-value> :gvec)
			 (<link-method> :gvec)
			 (<link-bdgs> :gvec))

	     (<imported-module> :gvec)

	     (<table-bucket> :gvec)
	     (<table> :abstract
		      (<symbol-table> :gvec)
		      (<string-table> :gvec)
		      (<string-ci-table> :gvec)
		      (<object-table> :gvec)
		      (<eq-table> :gvec))

	     (<list> :abstract 
		     (<pair> :gvec)
		     (<empty-list> :immob))))
|#


(define-class <gvec> (<object>) :gvec)
(define-class <bvec> (<object>) :bvec)

(define-class <dequeue> (<object>)
  (state type: <vector>)
  (front type: <fixnum> init-value: 0)
  (back type: <fixnum> init-value: 0))

;;;

(define-class <winding-contour> (<object>) :abstract)

;;;========================================================================
;;;  (used in `fluid.scm', but optimized in this files' functions)

(define-class <fluid-tl-contour> (<winding-contour>)
  ftlc-bindings
  ftlc-inside-values
  ftlc-saved-values)

;;;
;;;  an abstraction for where conditions might arise
;;;

(define-class <place> (<object>) :abstract)

;;; this subclass applies to both bytecoded functions (where the place
;;; code is the offset of the next BC instruction), and C-coded 
;;; functions (where the place code is just an index code)
;;;
;;; this is the object constructed by the make_function_place() function in
;;; entry.c

(define-class <function-place> (<place>)
  (template type: <template>)
  (function-place-code type: <fixnum>))
