#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/corelib/intrglue.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.11
 | Date:    1997.01.15 00:17:47
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: scheme-side interface for interrupts
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.11 %G% 00:17:47")


;;  this is an alist mapping RSSIG numbers
;;  to handler thunks

;;  this vector is available to the C interrupt dispatch code
;;  (runtime/intrs.c), from which it reads the handler proc

(define *interrupt-handlers* '#(#f #f #f #f #f #f #f #f #f #f #f #f 
				   #f #f #f #f #f #f #f #f #f #f #f 
				   #f #f #f #f #f #f #f #f #f #f #f 
				   #f #f #f #f #f #f #f #f #f #f #f
				   #f #f #f #f #f #f #f #f #f #f #f
				   #f #f #f #f #f #f #f #f #f #f #f
				   #f #f #f #f #f #f #f))

(%early-once-only
 (let loop ((i 0))
   (if (fixnum<? i 74)
       (begin
	 (vector-set! *interrupt-handlers* i default-handler)
	 (loop (add1 i))))))

;;
;; returns the RSSIG number
;;

(define-safe-glue (setup-signal-handler sig_name)
  literals: ('#(sigabrt sigfpe sigill sigint sigsegv sigterm))
{
  int n;

  /* ANSI C standard signals */
  static int signums[] = { SIGABRT, SIGFPE, SIGILL, SIGINT, SIGSEGV, SIGTERM };

  if (OBJ_ISA_FIXNUM(sig_name))
    n = fx2int(sig_name);
  else
    {
      obj v = LITERAL(0);
      unsigned i;

      n = -1;
      for (i=0; i < SIZEOF_PTR(v); i += SLOT(1))
	{
	  if (EQ(gvec_ref(v,i),sig_name))
	    {
	      n = signums[i/SLOT(1)];
	      break;
	    }
	}
      if (n < 0)
	scheme_error( "signal name '~s' not recognized", 1, sig_name );
    }
  if (signal( n, cast_to_signal_handler(c_signal_catcher) ) == SIG_ERR)
    {
      scheme_error( "signal id '~d' is invalid", 1, int2fx(n) );
    }
  REG0 = int2fx(RSSIG_C_SIGNAL + n);
  RETURN1();
})

(define-syntax (valid-rssig-num n)
  (let (((m <fixnum>) n))
    (if (and (fixnum>=? m 0)
	     (fixnum<? m (gvec-length *interrupt-handlers*)))
	m
	(error "invalid RSSIG number: ~s" n))))

;;
;; the proc should take a single argument which is the signal
;; instance information
;;

(define (register-interrupt-handler! handler-id (proc <function>))
  (let ((num (case handler-id
	       ((control-c) 0)
	       ((timer) 1)
	       ((data-ready) 2)
	       ((finalize) 3)
	       ((gc-flip) 4)
	       (else handler-id))))
    (if (not (fixnum? num))
	(error "~s: unrecognized interrupt handler id" 
	       handler-id))
    (gvec-set! *interrupt-handlers* (valid-rssig-num num) proc)))

(define (ignore-handler . args)
  )

(define (do-finalization lst)
  (let loop ((i lst))
    (if (pair? i)
	(begin
	  (finalize (car i))
	  (loop (cdr i))))))

(%early-once-only
 (register-interrupt-handler! 'gc-flip ignore-handler)
 (register-interrupt-handler! 'finalize do-finalization))


(define (default-handler . args)
  (process-abort "No handler for signal" args))

;;
;;  a pointer to this template is scheme_global[6], and is used
;;  by runtime/intrs.c to be the continuation from an interrupt
;;
;;

(define-glue (continue-intr-template) :template
{
jump_addr f;

    RESTORE_CONT0();	/* personally, we don't have anything
    			   saved in our continuation */
    f = half_restore();			/* but get the interrupted
    					   context's fixed registers */
    arg_count_reg = restore_arb();	/* and all the other regs */
    os_set_sigenable( YES );

#ifdef GNU_VINSNS
    goto *f;
#else
    return f;
#endif
})

;;;  this template is special in the following way:
;;;  it's entry point is never exercised.  The runtime system
;;;  (entry.c) constructs a continuation to it's monotone[1]
;;;  which is responsible for escaping the quasi-interpretive
;;;  spin loop when a call to call_scheme() is complete
;;;
;;;  although the first monotone is never used, I nevertheless
;;;  constructed it so that if this function were called, it
;;;  would terminate the call to call_scheme()

(define-glue (return-from-call-scheme) :template
{
  JUMP(arg_count_reg,finish_run_1);  /* never used */
}
("finish_run_1" {
   done_w_call_scheme();
   RETURN1(); /* never reached */
}))

(define-glue (enable-subprocess-capture)
{
  enable_subprocess_capture();
  RETURN0();
})

#|
(define-glue (test-call-scheme proc arg)
{
   REG0 = call_scheme( proc, 1, arg );
   RETURN1();
})
|#
