#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/corelib/keywords.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.4
 | Date:    1997.01.08 18:22:15
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: keyword/flag handling (symbols that end/start with `:')
 |------------------------------------------------------------------------|
 | Notes:
 |     Currently, "keywords" are just symbols with a : on the
 |     end, but they may get their own class at some point
 |     
 |     this is included by compiler/util/keywords.scm
 `------------------------------------------------------------------------|#


(define (keyword? x)
  (and (symbol? x)
       (let* (((s <string>) (symbol->string x))
	      ((n <fixnum>) (string-length s)))
	 (and (fixnum>? n 1)
	      (eq? (bvec-ref s (sub1 n)) 58))))) ;; `58' is the code for #\:

(define (flag? x)
  (and (symbol? x)
       ;; all strings are bvecs of length at least 1 (the terminating NUL
       ;; char), so we can reference byte 0 unconditionally
       (eq? (bvec-ref (symbol->string x) 0) 58)))

(define (keyword->symbol x)
  (assert (keyword? x))
  (let ((s (symbol->string x)))
    (string->symbol (substring s 0 (sub1 (string-length s))))))

(define (sa/as-strings items)
  (cond
   ((null? items)
    '())
   ((string? (car items))
    (cons (car items) (sa/as-strings (cdr items))))
   ((symbol? (car items))
    (cons (symbol->string (car items))
	  (sa/as-strings (cdr items))))
   ((fixnum? (car items))
    (cons (fixnum->string (car items) 10)
	  (sa/as-strings (cdr items))))
   (else
    (error "symbol-append: invalid component `~s'" (car items)))))
	  
(define (symbol-append . items)
  (string->symbol (apply* (sa/as-strings items) string-append)))

(define (symbol->keyword (s <symbol>))
  (symbol-append s ":"))

(define (flag->symbol x)
  (assert (flag? x))
  (let ((s (symbol->string x)))
    (string->symbol (substring s 1))))

(define (symbol->flag (s <symbol>))
  (symbol-append ":" s))
