#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/corelib/macros.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.8
 | Date:    1996.12.18 16:10:46
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: standard scheme macros
 `------------------------------------------------------------------------|#

;;(file-version-string "%Z%%W% 1.8 %G% 16:10:46")

;; 
;;
;;  macros to support standard scheme
;;

(define-syntax cond
  (syntax-form (('else . action))
    (begin . action))
  (syntax-form ((guard) . more-clauses)
    (let ((temp guard)) 
      (if temp 
	  temp 
	  (cond . more-clauses))))
  (syntax-form ((guard '=> proc) . more-clauses)
    (let ((temp guard)) 
      (if temp 
	  (proc temp)
	  (cond . more-clauses))))
  (syntax-form ((guard . actions) . more-clauses)
    (if guard 
	(begin . actions) 
	(cond . more-clauses)))
  (syntax-form ()
    (begin)))

(define-syntax or
  (syntax-form (clause)
    clause)
  (syntax-form (clause . more-clauses)
    (let ((temp clause))
      (if temp
	  temp
	  (or . more-clauses))))
  (syntax-form () #f))

(define-syntax cons*
  (syntax-form (a b)
    (cons a b))
  (syntax-form (a b . more)
    (cons a (cons* b . more))))

(define-syntax and
  (syntax-form (clause1 clause2 . more-clauses)
    (if clause1
	(and clause2 . more-clauses) 
	#f))
  (syntax-form () 
    #t)
  (syntax-form (clause) 
    clause))

(define-syntax case
  (syntax-form (expr . branches)
    (letrec-syntax
	((test-cases
	  (syntax-form (var ((g0 . g1) . actions) . more-cases)
	    (if (test-guards var g0 . g1)
		(begin . actions)
		(test-cases var . more-cases)))
	  (syntax-form (var ('else . actions))
	    (begin . actions))
	  (syntax-form (var)
	    (begin)))
	 (test-guards
	  (syntax-form (var item)
	    (eq? var (mquote item)))
	  (syntax-form (var item . more-items)
	    (if (eq? var (mquote item))
		#t
		(test-guards var . more-items)))
	  (syntax-form (var)
	    #f)))
      (let ((temp expr))
	(test-cases temp . branches)))))

(define-rewriter (do form)
  (let ((bdgs (car (cdr form)))
	(test (car (cdr (cdr form))))
	(body (cdr (cdr (cdr form))))
	(loop-name (gensym)))
    (let ((vars (map car bdgs))
	  (inits (map cadr bdgs))
	  (incs (map (lambda (x) 
		       (if (null? (cdr (cdr x)) )
			   (car x) 
			   (car (cdr (cdr x))) ))
		     bdgs)))
      (list
       'let
       loop-name 
       (map list vars inits)
       (list
	'if
	(car test)
	(cons 'begin (cdr test))
	(list
	 'begin
	 (cons 'begin body)
	 (cons loop-name incs)))))))


(define-rewriter (let* form)
  (let ((bdgs (car (cdr form))))
    (if (null? bdgs)
	(cons 'begin (cdr (cdr form)))
	(list
	 'let
	 (list (car bdgs))
	 (cons
	  'let*
	  (cons (cdr bdgs)
		(cdr (cdr form))))))))

;;
;;  this rewriter uses the contained macro-pattern->matcher procedure
;;  to implement `define-macro' in terms of `define-rewriter'
;;
;;  basically, a pattern-matching macro definition like
;;
;;   (define-macro (foo x y) `(cons ,x ,y))
;;
;;  becomes a rewriter
;;
;;   (define-rewriter (foo form)
;;     (let ((x '#uninit)
;;           (y '#uninit))
;;       <validate form against pattern (x y) and set x and y to
;;        appropriate s-expr values>
;;       `(cons ,x ,y)))
;;

(define-rewriter (define-macro form)
  ;;
  ;;  this procedure is used to traverse a macro pattern
  ;;  and construct a program which will, at runtime (which,
  ;;  since this is code for a rewriter macro, will be at compile
  ;;  time; specifically, a use of the macro being defined),
  ;;  validate the presence of the pattern and assign the pattern
  ;;  variables the correct values
  ;;
  (define (macro-pattern->matcher pat xpr vars)
    (cond
     ((symbol? pat) 
      (vector-set! vars 0 (cons pat (vector-ref vars 0)))
      `(set! ,pat ,xpr))
     ((pair? pat)
      (let ((v1 (gensym))
	    (v2 (gensym)))
	`(if (pair? ,xpr)
	     (let ((,v1 (car ,xpr))
		   (,v2 (cdr ,xpr)))
	       ,(macro-pattern->matcher (car pat) v1 vars)
	       ,(macro-pattern->matcher (cdr pat) v2 vars))
	     (error "arg ~s doesn't match pattern ~s"
		    ,xpr ',pat))))
     ((null? pat)
      `(if (null? ,xpr)
	   #t
	   (error "excess arg ~s" ,xpr)))
     (else
      (error "illegal macro pattern: ~s" pat))))
  ;;
  ;;  the actual transformer proc
  ;;
  (define (defmacro-xformer form)
    (let ((name (caadr form))
	  (pattern (cdadr form))
	  (body (cddr form))
	  (vars (vector '())))
      (let ((pat (macro-pattern->matcher pattern '(cdr $form) vars)))
	(let ((c `(define-rewriter (,name $form)
		    (let (,@(map (lambda (v)
				   (list v ''#uninit))
				 (vector-ref vars 0)))
		      ,pat
		      ,@body))))
	  ;; debugging: (pprint c)
	  c))))
  ;;
  ;;  call the transformer proc
  ;;  (it's factored out this way for easier testing; you can take the
  ;;  xformer def'n outside to the top level and hand it expressions
  ;;  and see what it gets back)
  ;;
  (defmacro-xformer form))

(define-macro (define-inline (proc . args) . body)
  `(define-syntax ,proc
     (else (lambda (,@args) ,@body))))
