#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/editinp/editinp.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.10
 | Date:    1996.12.19 13:09:12
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: Friendly input, with prompts & all
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.10 %G% 13:09:12")

;;
;;  a prompting input port.
;;  this is where the readline library gets plugged in
;;

(define-class <edit-input-port> (<input-port>)
  filler-proc
  primary-prompt
  secondary-prompt
  completions
  buffer
  buffer-index
  use-secondary?)

(define (fill-edit-input-port (self <edit-input-port>))
  ((filler-proc self) self))

(define (open-edit-port)
  (make <edit-input-port>
	filler-proc: (if (and (readline-enabled?)
			      (eq? (current-input-port) 
				   $standard-input-port))
			 (console-prompting-filler)
			 (basic-prompting-filler (current-input-port)
						 (current-output-port)))
	primary-prompt: "? "
	secondary-prompt: " "
	completions: '()
	buffer: ""
	buffer-index: 1
	use-secondary?: #f))

;;
;; cause some operations on non-edit ports to be ignored
;;

(define-method set-primary-prompt! ((self <input-port>) arg))
(define-method set-completions! ((self <input-port>) arg))
;;

;;
;;  if this <edit-input-port> is attached to the "console" (stdin),
;;  then we can use our readline subsystem
;;

(define (console-prompting-filler)
  (lambda ((self <edit-input-port>))
    (let ((line (readline-read-line 
		 (completions self) 
		 (if (use-secondary? self)
		     (secondary-prompt self)
		     (primary-prompt self)))))
      (if line
	  (begin
	    (readline-add-to-history line)
	    (set-buffer! self line))
	  (begin
	    (set-buffer! self $eof-object)))
      (set-buffer-index! self 0))))

;;
;; if we're connected to some other port, then just do a basic
;; prompt/read thing

(define (basic-prompting-filler raw-input-port raw-output-port)
  (lambda ((self <edit-input-port>))
    (display (if (use-secondary? self)
		 (secondary-prompt self)
		 (primary-prompt self))
	     raw-output-port)
    (flush-output-port raw-output-port)
    (set-buffer! self (read-line raw-input-port))
    (set-buffer-index! self 0)))

(define-method input-port-read-char ((self <edit-input-port>))
  (let (((ix <fixnum>) (buffer-index self)))
    (if (eof-object? (buffer self))
	(begin
	  (set-buffer! self "")
	  (set-buffer-index! self 1)
	  $eof-object)
	(if (< ix (string-length (buffer self)))
	    (begin
	      (set-buffer-index! self (+ ix 1))
	      (string-ref (buffer self) ix))
	    (if (eq? ix (string-length (buffer self)))
		(begin
		  (set-buffer-index! self (+ ix 1))
		  #\newline)
		(begin
		  (fill-edit-input-port self)
		  (input-port-read-char self)))))))


(define-method input-port-peek-char ((self <edit-input-port>))
  (let (((ix <fixnum>) (buffer-index self)))
    (if (eof-object? (buffer self))
	$eof-object
	(if (< ix (string-length (buffer self)))
	    (string-ref (buffer self) ix)
	    (if (eq? ix (string-length (buffer self)))
		#\newline
		(begin
		  (fill-edit-input-port self)
		  (input-port-peek-char self)))))))

  

;;

(define (scanner-for-edit-port port)
  (let ((meth (find-method input-port-scan-token (list port)))
	(p port) ;; copy of port for scanner->port to use
	(ss set-use-secondary?!))
    (lambda ()
      (if ss
	  (bind ((type val (meth p)))
	    (ss p #t)
	    (set! ss #f)
	    (values type val))
	  (meth p)))))

(define-method input-port-read ((self <edit-input-port>))
  (let ((result (read:parse-object (scanner-for-edit-port self))))
    (set-use-secondary?! self #f)
    result))
