#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/mathlib/arith.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.7
 | Date:    1996.09.12 11:08:47
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: high-level arithmetic (transcendentals, etc)
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.7 %G% 11:08:47")


(define (bad-arg proc args)
    (error "~s: Bad argument in: ~s" proc args))

#|
   nobody uses these, and there are no methods on them, so remove
   them by default...

(define-generic-function binary+)
(define-generic-function binary-)
(define-generic-function binary*)
(define-generic-function binary/)
(define-generic-function unary/)
(define-generic-function unary-)
|#

#|
(define-glue (quick-binary+ x y) literals: ((& binary+))
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = FX_ADD(x,y);
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( fx2int(x) + extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = make_float( extract_float(x) + fx2int(y) );
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( extract_float(x) + extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else
	goto invoke_generic_fn;
    RETURN(1);

invoke_generic_fn:
    INVOKE_GENERIC();
})

(define-glue (quick-binary- x y) literals: ((& binary-))
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = FX_SUB(x,y);
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( fx2int(x) - extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = make_float( extract_float(x) - fx2int(y) );
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( extract_float(x) - extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else
	goto invoke_generic_fn;
    RETURN(1);

invoke_generic_fn:
    INVOKE_GENERIC();
})

(define-glue (quick-binary* x y) literals: ((& binary*))
{
    if (OBJ_ISA_FIXNUM(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = FX_MUL(x,y);
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( fx2int(x) * extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else if (LONGFLOAT_P(x))
    {
	if (OBJ_ISA_FIXNUM(y))
	    REG0 = make_float( extract_float(x) * fx2int(y) );
	else if (LONGFLOAT_P(y))
	    REG0 = make_float( extract_float(x) * extract_float(y) );
	else
	    goto invoke_generic_fn;
    }
    else
	goto invoke_generic_fn;
    RETURN(1);

invoke_generic_fn:
    INVOKE_GENERIC();
})

(define-glue (quick-binary/ x y) literals: ((& binary/))
{
double rx, ry;

    if (OBJ_ISA_FIXNUM(x))
	rx = fx2int(x);
    else if (LONGFLOAT_P(x))
	rx = extract_float(x);
    else
	goto invoke_generic_fn;
	
    if (OBJ_ISA_FIXNUM(y))
	ry = fx2int(y);
    else if (LONGFLOAT_P(y))
	ry = extract_float(y);
    else
	goto invoke_generic_fn;

    REG0 = make_float( rx / ry );
    RETURN(1);

invoke_generic_fn:
    INVOKE_GENERIC();
})

(define-glue (quick-unary/ x) literals: ((& unary/))
{
double rx;

    if (OBJ_ISA_FIXNUM(x))
	rx = fx2int(x);
    else if (LONGFLOAT_P(x))
	rx = extract_float(x);
    else
	INVOKE_GENERIC();

    REG0 = make_float( 1.0 / rx );
    RETURN(1);
})

(define-glue (quick-unary- x) literals: ((& unary-))
{
    if (OBJ_ISA_FIXNUM(x))
	REG0 = int2fx(-fx2int(x));
    else if (LONGFLOAT_P(x))
	REG0 = make_float( -extract_float(x) );
    else
	INVOKE_GENERIC();
    RETURN(1);
})
|#

(define-method floor ((self <fixnum>))
  self)

(define-method floor ((self <double-float>))
  (float-floor self))

;;

(define-method ceiling ((self <fixnum>))
  self)

(define-method ceiling ((self <double-float>))
  (float-ceiling self))

;;

(define-method truncate ((self <fixnum>))
  self)

(define-method truncate ((self <double-float>))
  (raw-int->double-float (float-truncate self)))

(define-method round ((self <fixnum>))
  self)

(define-method round ((self <double-float>))
  (float-round self))

(define-glue (float-round x)
{
  IEEE_64 a = extract_float(x);
  INT_32 ai;
    
  if (a < 0)
    {
      ai = (int)(a - 0.5);
      if (ai >= -536870912)
	REG0 = int2fx( ai );
      else
	REG0 = x;
    }
  else
    {
      ai = (int)(a + 0.5);
      if (ai <= 536870911)
	REG0 = int2fx( ai );
      else
	REG0 = x;
    }
    RETURN(1);
})

(define-method exact->inexact ((self <fixnum>))
  (raw-int->double-float self))

(define-method exact->inexact ((self <double-float>))
  self)

(define-method inexact->exact ((self <fixnum>))
  self)

(define-method inexact->exact ((self <double-float>))
  (float-truncate self))

#|
(define-rewriter (arithmetic-order-predicate form)
  (let* ((op (symbol->string (cadr form)))
	 (gf (string->symbol (string-append "binary" op "?")))
	 (fx (if (eq? (cadr form) '=)
		 'eq?
		 (string->symbol (string-append "fixnum" op "?"))))
	 (fl (string->symbol (string-append "float" op "?"))))
    `(begin
       (define-method ,gf ((a <fixnum>) b)
	 (if (fixnum? b)
	     (,fx a b)
	     (,fl (raw-int->raw-float a) b)))
       
       (define-method ,gf ((a <double-float>) b)
	 (if (double-float? b)
	     (,fl a b)
	     (if (fixnum? b)
		 (,fl a (raw-int->raw-float b))
		 (,fl a (exact->inexact b))))))))

(arithmetic-order-predicate <)
(arithmetic-order-predicate <=)
(arithmetic-order-predicate >)
(arithmetic-order-predicate >=)
(arithmetic-order-predicate =)
|#

(define-rewriter (real-operator form)
  (let* ((op (symbol->string (cadr form)))
	 (fl (string->symbol (string-append "float-" op))))
    `(define (,(cadr form) (z <number>))
       (if (double-float? z)
	   (,fl z)
	   (,fl (exact->inexact z))))))

(real-operator exp)
(real-operator log)
(real-operator sin)
(real-operator cos)
(real-operator tan)
(real-operator asin)
(real-operator acos)
(real-operator sqrt)

;;

(define-syntax with-raw-float 
  (syntax-form (head x)
    (if (double-float? x)
	(head x)
	(if (fixnum? x)
	    (head (raw-int->raw-float x))
	    (head (exact->inexact x)))))
  (syntax-form (head x y)
    (if (double-float? x)
	(if (double-float? y)
	    (head x y)
	    (if (fixnum? y)
		(head x (raw-int->raw-float y))
		(head x (exact->inexact x))))
	(if (fixnum? x)
	    (if (double-float? y)
		(head (raw-int->raw-float x) y)
		(head (raw-int->raw-float x) (exact->inexact x)))
	    (head (exact->inexact x)
		  (exact->inexact y))))))

(define (atan (z <number>) . rest)
  (if (null? rest)
      (with-raw-float float-atan1 z)
      (if (null? (cdr rest))
	  (with-raw-float float-atan2 z (car rest))
	  (error "atan: too many args"))))

(define (float-result-expt z1 z2)
  (with-raw-float float-pow z1 z2))
