#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/mlink/linkload.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.8
 | Date:    1996.07.30 22:49:07
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: locate, load & bind modules
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.8 %G% 22:49:07")

;;
;;  dynamically bind a module
;;

(define *boot-modules* '()) ;; filled in at boot image linkage time
(define *space* #f)

(define (installed-modules)
  (or *space* *boot-modules*))

(define (install-module! name module)
  (set! *space* (cons (cons name module) (installed-modules))))

;;
;;  `link-load-module' is only used to load modules for execution
;;  if you are doing different things (like the to-C-compiler), then
;;  you would use load-module directly, because you need to manage
;;  your own module namespace and not finalize/patch the loaded modules
;;

(define (link-load-module name path)
  (let ((m (load-module name path #t)))
    (link-into name m (installed-modules) #f)
    (for-each finalize-class (module-classes m))
    (for-each finalize-generic-function (module-generic-functions m))
    (for-each (lambda (method-set)
		(patch-implicit-methods (cdr method-set)
					(car method-set)))
	      (module-implicit-methods m))
    (install-module! name m)
    (for-each (lambda (thunk)
		(thunk))
	      (init-thunks m))
    (let ((startup-thunks (vector-ref (rscheme-global-ref 0) 3)))
      (append! startup-thunks (init-thunks m)))
    m))

(define *module-search-path* '())

(define (module-search-path)
  *module-search-path*)

(define (set-module-search-path! lst)
  (set! *module-search-path* lst))

(define (push-module-search-path! (p <directory-name>))
  (set! *module-search-path* (cons p *module-search-path*)))

(%early-once-only
 (push-module-search-path! (string->dir "~/lib/rs/modules"))
 (push-module-search-path! (string->dir "[resource]/modules")))

(define (get-module (name <symbol>))
  (let ((m (assq name (installed-modules))))
    (if m
	(cdr m)
	(let ((path (basic-locate (symbol->string name)
				  (module-search-path)
				  '("mif"))))
	  (if path
	      (link-load-module name path)
	      (error "~s: couldn't find module" name))))))


;; most bindings issue a warning and use the new one

(define-method resolve-import-conflict ((self <binding>)
					(new-bdg <binding>)
					(name <symbol>)
					(tle <top-level-contour>))
  (format #t "warning: ~s already bound to ~s\n" name self)
  (table-insert! (table tle) name new-bdg))

(define-method resolve-import-conflict ((self <top-level-var>)
					(new-bdg <binding>)
					(name <symbol>)
					(tle <top-level-contour>))
  (if (instance? new-bdg <top-level-var>)
      (begin
	(if (not (or (eq? (value self) '#unbound)
		     (eq? (value self) (value new-bdg))))
	    (format #t "warning: TLV ~s already bound to ~s\n" 
		    name (value self)))
	(set-value! self (value new-bdg)))
      (begin
	(format #t "warning: ~s already bound to ~s = ~s\n" 
		name self (value self))
	(table-insert! (table tle) name new-bdg))))

#|
  other possible `resolve-import-conflict' methods...
  
	     (if (and (instance? b <top-level-var>)
		      (instance? v <top-level-var>))
		 (set-value! b (value v))
		 (if (and (instance? b <primop>)
			  (instance? v <primop>)
			  (eq? (name b) (name v)))
		     #t ;; do nothing
|#

;;

(define (use-module-in mname m envt)
  (let ((t (table envt)))
    (set-dirty?! envt #t)
    (table-for-each
     (module-exports m)
     (lambda (h k v)
       (let ((b (table-lookup t k)))
	 (if b
	     (resolve-import-conflict b v k envt)
	     (table-install! t h k v)))))
    (values)))

(define (use-in name envt)
  (if (eq? name 'all)
      (for-each (lambda (ent)
		  (use-module-in (car ent) (cdr ent) envt))
		(installed-modules))
      (use-module-in name (get-module name) envt)))
