#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/mlink/mifio.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.12
 | Date:    1996.09.11 22:21:06
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: <module> image file (.mif) I/O
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.12 %G% 22:21:06")

;; load and save module interfaces

(define *mif-load-defs* #f)
(define *mif-save-defs* #f)
(define *mifio-defs* '())

(define (add-mifio-class name class)
  (set! *mifio-defs* (cons (cons name class) *mifio-defs*))
  (set! *mif-load-defs* #f)
  (set! *mif-save-defs* #f))
  
(define-rewriter (mifio-class form)
  (let ((name (cadr form))
	(class (caddr form)))
    `(%early-once-only (add-mifio-class ,name ,class))))

(define (mif-load-defs)
  (if (not *mif-load-defs*)
      (let ((t (make-table string=? string->hash)))
	(set! *mif-load-defs* t)
	(for-each
	 (lambda (e)
	   (table-insert! *mif-load-defs*
			  (car e)
			  (cdr e)))
	 *mifio-defs*)))
  *mif-load-defs*)

(define (mif-save-defs)
  (if (not *mif-save-defs*)
      (set! *mif-save-defs* (list->vector *mifio-defs*)))
  *mif-save-defs*)

(define (mif-table-patching-load-defs)
  (let ((x (make-table string=? string->hash)))
    (for-each
     (lambda (e)
       (table-insert! x (car e) (cdr e)))
     *mifio-defs*)
    ;;
    ;; when we load an old-tables module, the loaded symbol tables 
    ;; will be in 0.5
    ;; format, so make 'em a different class so we can recognize
    ;; them and fix em.
    ;;
    (table-insert! x "<symbol-table>" <0.5-symbol-table>)
    x))

;;
;; methinks this is somewhat Linux specific...
;;

(define (find-loadable-c-module (path <file-name>) (cm <string>))
  (let ((p (pathname->os-path (make <file-name>
				    filename: (string-append "lib" cm)
				    extension: "so"
				    file-directory: (file-directory path)))))
    (if (os-file-exists? p)
	p
	#f)))

(define (ensure-c-modules-loaded why (path <file-name>) mx)
  (for-each 
   (lambda (cm)
     (if (not (find-linked-module cm))
	 ;; not already loaded... see if we can load it
	 (let ((loadable-path (find-loadable-c-module path cm)))
	   (if loadable-path
	       (begin
		 (format #t "loading C module ~a: ~a\n" cm loadable-path)
		 (load-c-module loadable-path cm))
	       (error "~a: required C module `~a' not loaded" why cm)))))
   (caddr mx)))

(define (load-module (m-name <symbol>) 
		     (path <file-name>)
		     executable?)
  ;;
  ;;  load and exercise the MX file (meta info)
  ;;
  (let ((mx-info (call-with-input-file (pathname->os-path 
					(extension-related-path path "mx"))
		   read)))
    ;;
    (if executable?
	(ensure-c-modules-loaded m-name path mx-info))
    ;;
    ;;  load the actual image file
    ;;
    (load-image (pathname->os-path path)
		(mif-load-defs)
		(if executable?
		    #f
		    <part-descr>))))

(define (save-module path m)
  (save-image path m (mif-save-defs)))

(mifio-class "<<class>>" <<class>>)
(mifio-class "<symbol>" <symbol>)
(mifio-class "<closure>" <closure>)
(mifio-class "<vector>" <vector>)
(mifio-class "<string>" <string>)
(mifio-class "<pair>" <pair>)
(mifio-class "<double-float>" <double-float>)
(mifio-class "<table>" <table>)
(mifio-class "<string-table>" <string-table>)
(mifio-class "<string-ci-table>" <string-ci-table>)
(mifio-class "<object-table>" <object-table>)
(mifio-class "<eq-table>" <eq-table>)
(mifio-class "<table-bucket>" <table-bucket>)


;;
;;  these are built in to 0.6, but not 0.5

(mifio-class "<top-level-var>" <top-level-var>)
(mifio-class "<method>" <method>)
(mifio-class "<<standard-class>>" <<standard-class>>)
(mifio-class "<top-level-contour>" <top-level-contour>)

;; note that a <lexical-contour> can occur in a module image if
;; a <macro> captures a contour containing macros.  This may be
;; a good reason right here to distinguish variable-type contours
;; from syntactic contours, since the latter preserve top-levelness
;; and can be in an image, wheras the former do not.

(mifio-class "<lexical-contour>" <lexical-contour>)

(mifio-class "<generic-function>" <generic-function>)
(mifio-class "<binding-envt>" <binding-envt>)
(mifio-class "<slot-descriptor>" <slot-descriptor>)

(mifio-class "<rewriter>" <rewriter>)
(mifio-class "<macro>" <macro>)
(mifio-class "<macro-form>" <macro-form>)
(mifio-class "<substitution>" <substitution>)
(mifio-class "<primop>" <primop>)
