#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/objsys/genericf.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.8
 | Date:    1997.01.23 09:08:05
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: generic function dispatcher and find-method
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.8 %G% 09:08:05")

#|
   This is the default generic function dispatcher

   It's what gets called when a generic function is
   invoked.  The environment, as set up by the 
   define-generic-function compiler, consists (initially)
   of a single frame, which contains a pointer to
   the <generic-function> instance (a back-pointer to 
   the function, since the function ptr is lost in our
   calling convention), and a slot with initialize value ()
   which may be used for caching or whatever.
|#

(define-safe-glue (get-gf-lru-histogram resetq)
{
  REG0 = collect_gf_cache_histogram( truish(resetq) );
  RETURN1();
})

(define-glue (generic-function-dispatch) :template
  literals: ((& load-cache-and-call)
	     (& call-gf-with-zero-args))
{
obj c, m, impl;

    USE_FUNCTION_ENVT();
    if (arg_count_reg < 1)
    {
        REG0 = LEXREF0(0);
	APPLY( 1, TLREF(1) );
    }
    else
    {
     c = object_class(REG0);
#if ACCUM_GF_CACHE_HISTOGRAM
     {
     obj prev = ZERO, first = LEXREF0(1);
     extern int gf_cache_hit_depth[100];
     int lruposn;

     for (lruposn=0,m=first; !NULL_P(m); prev=m,m=pair_cdr(m),lruposn++)
       {
	 assert( PAIR_P(m) );
	 impl = pair_car(m);
	 if (EQ(pair_car(impl),c))
           { 
             if (!EQ(prev,ZERO))
               {
                 pair_set_cdr( prev, pair_cdr(m) );
                 pair_set_cdr( m, first );
                 LEXSET0(1,m);
               }
             if (lruposn < 100)
                gf_cache_hit_depth[lruposn]++;
	     APPLY( arg_count_reg, pair_cdr(impl) );
           }
       }
     }
#else
     for (m=LEXREF0(1); !NULL_P(m); m=pair_cdr(m))
       {
	 assert( PAIR_P(m) );
	 impl = pair_car(m);
	 if (EQ(pair_car(impl),c))
	   APPLY( arg_count_reg, pair_cdr(impl) );
       }
#endif
     COLLECT0();
     REG1 = REG0;
     REG0 = LEXREF0(0);
     APPLY( 2, TLREF(0) );
   }
})

(define-glue (pre-compute-dispatch)
  literals: ((& load-cache)
	     (& call-gf-with-zero-args))
{
obj c, m, impl;

    /* adopt the environment of the GF we're examining */
    envt_reg = gvec_ref( REG0, SLOT(0) );

    if (arg_count_reg < 2)
    {
        REG0 = LEXREF0(0);
	APPLY( 1, TLREF(1) );
    }
    else
    {
     c = object_class(REG1);

     for (m=LEXREF0(1); !NULL_P(m); m=pair_cdr(m))
       {
	 assert( PAIR_P(m) );
	 impl = pair_car(m);
	 if (EQ(pair_car(impl),c))
	   {
	     REG0 = pair_cdr(impl);
	     RETURN1();
	   }
       }
     COLLECT1();
     APPLY( 2, TLREF(0) );
   }
})

(define-syntax (std-generic-function-name gf)
  (gvec-ref gf 4))

(define (call-gf-with-zero-args (gf <generic-function>))
  (error "~s: generic function called with no arguments"
	 (std-generic-function-name gf)))

(define (does-not-understand (gf <generic-function>) (args <list>))
  (error "~s: generic function doesn't understand ~s\n(rest are ~s)"
	 (std-generic-function-name gf)
	 (car args) (cdr args)))

(define (load-cache-and-call (gf <generic-function>) (args <pair>))
  (apply* args (load-cache gf args)))

(define (load-cache (gf <generic-function>) (args <pair>))
  (let ((m (find-method gf args)))
    (if m
	;;
	;; store it in the cache..
	;;
	(let ((gfe (gvec-ref gf 0)))
	  (gvec-set! gfe 
		     2
		     (cons (cons (object-class (car args)) m)
			   (gvec-ref gfe 2)))
	  m)
	(does-not-understand gf args))))

(define-syntax (method-key-class m)
  (car (gvec-ref m 2)))

(define-syntax (std-generic-function-methods gf)
  (gvec-ref gf 2))

(define-syntax (method-function-specializers m)
  (gvec-ref m 2))


(define (find-method (gf <generic-function>) (args <pair>))
  (let ((key-class (object-class (car args))))
    (let loop ((i (std-generic-function-methods gf)))
      (if (pair? i)
	  (if (subclass? key-class (method-key-class (car i)))
	      (car i)
	      (loop (cdr i)))
	  #f))))
