#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/objsys/metaobj.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.8
 | Date:    1996.12.17 15:43:07
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: internal class protocol implementation
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.8 %G% 15:43:07")


(define (compute-class-precedence-list (c <<class>>))
  (if (null? (superclasses c))
      (list c)
      (cons c (compute-class-precedence-list (car (superclasses c))))))


;; return a list of all the class's <slot-descriptor>'s (including inherited)

(define (slot-descriptors (c <<class>>))
  (if (null? (superclasses c))
      (class-direct-slots c)
      (append (slot-descriptors (car (superclasses c)))
	      (class-direct-slots c))))


(define (class-instance-size (c <<class>>))
  (gvec-ref c 6))

(define-method name ((self <<class>>))
  (class-name self))

(define (class-superclasses (class <<class>>))
  (gvec-ref class 3))

(define (class-direct-slots (class <<class>>))
  (gvec-ref class 4))

;; find the <slot-descriptor> for a particular slot (by slot index)
;; in a particular class
;;
;; returns #f if the <slot-descriptor> is not found

(define (slot-descriptor-for-class-slot (class <<class>>)
					(slot <fixnum>))
  (let loop ((a (gvec-ref class 4)))
    (if (null? a)
	(if (pair? (class-superclasses class))
	    (slot-descriptor-for-class-slot (car (class-superclasses class))
					    slot)
	    #f)
	(if (eq? (gvec-ref (car a) 4) slot)
	    (car a)
	    (loop (cdr a))))))

;; returns, computing if necessary, the list of all slot descriptors
;; for the given class

(define (class-compute-slots (class <<class>>))
  (or (gvec-ref class 5)
      (let* ((sup (class-superclasses class))
	     (s (append (class-direct-slots class)
			(if (null? sup)
			    '()
			    (class-compute-slots (car sup))))))
	(gvec-set! class 5 s)
	s)))

;;

(define-method slot-type ((self <slot-descriptor>))
  (type-restriction self))

(define-method slot-value ((self <slot-descriptor>) item)
  (if (memq self (slot-descriptors (object-class item)))
      (gvec-ref item (index self))
      (error "slot-value: slot ~s not present in ~s" (name self) item)))

(define-method set-slot-value! ((self <slot-descriptor>) item value)
  (if (memq self (slot-descriptors (object-class item)))
      (if (instance? value (type-restriction self))
	  (gvec-set! item (index self) value)
	  (error "set-slot-value!: value ~s for slot ~s in ~s is\nnot compatible with the type restriction"
		 value (name self) item))
      (error "set-slot-value!: slot ~s not present in ~s" (name self) item)))
