#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/objsys/methods.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.5
 | Date:    1996.06.26 22:14:51
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: code to finish creating GFs and classes
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.5 %G% 22:14:51")

;; class introspection

(define (finalize-class (c <<class>>))
  ;(format #t "finalizing class: ~s\n" (name c))
  ;; compute the class precedence list
  (set-class-precedence-list! 
   c
   (compute-class-precedence-list c)))
  ;(format #t "  CPL: ~s\n" (class-precedence-list c))


(define (finalize-generic-function (gf <generic-function>))
  ;(format #t "finalizing generic function: ~s\n" (generic-function-name gf))
  (set-template! gf generic-function-dispatch))

;; note: the method may be specialized on classes
;; which are extern (ie, referred to through patches)

(define (target-add-method (gf <generic-function>)
			   (method <method>))
  (assert (instance? gf <generic-function>))
  (assert (instance? method <method>))
  ;;
  ;; clear it's cache
  ;;
  (gvec-set! (environment gf) 2 '())
  ;;
  (let ((methods (generic-function-methods gf)))
    ;; 
    ;; examine the function's specializers to see where to add it
    ;;
    (let-syntax ((insert-before (syntax-form (prev rest)
				  (if prev
				      (set-cdr! prev (cons method rest))
				      (set-generic-function-methods! gf
						       (cons method rest)))
				  method))
		 (method-key-class (syntax-form (m)
				      (car (function-specializers m)))))
      (let loop ((i methods)
		 (prev #f))
	(cond ((null? i)
	       (insert-before prev '()))
	      ((eq? (method-key-class method)
		    (method-key-class (car i)))
	       ; this is a REPLACEMENT of an existing method
	       (set-car! i method)
	       method)
	      ((subclass? (method-key-class method)
				 (method-key-class (car i)))
	       (insert-before prev i))
	      (else (loop (cdr i) i)))))))

;;
;;  `patch-implicit-methods'
;;
;;  this function goes through a list of methods that are being
;;  loaded (as from a module) and installs the appropriate template
;;  pointer.
;;
;;  normal methods already have a template pointer; this function
;;  is used to patch up those that use "well known" templates, such
;;  as getter and setter methods.
;;
;;  (I wonder if this should be handled by using named pointers in
;;   the context of module unpickling...  Would cross-compilation
;;   be a problem?)

(define (patch-implicit-methods (methods <list>) implicit-form)
  (let ((t (case implicit-form
	     ((getter-template) getter-template)
	     ((setter-template) setter-template)
	     (else (error "implicit method form `~s' not recognized"
			  implicit-form)))))
    (for-each (lambda ((meth <method>))
		(set-template! meth t))
	      methods)))

