#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/paths/rootpath.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.4
 | Date:    1996.06.29 21:07:17
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: Manage the roots of absolute pathnames
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.4 %G% 21:07:17")

(define $system-root #f)    ;; /
(define $user-root #f)      ;; ~
(define $install-root #f)   ;; [install]
(define $resource-root #f)  ;; [resource]

(%early-once-only
 (set! $system-root (make <root-dir> 
			  root-name: "/"
			  expanded-name: "/"))
 (set! $user-root (make <root-dir> 
			root-name: "~/"
			expanded-name: "~/"))
 (set! $install-root (make <root-dir> 
			   root-name: "[install]/"
			   expanded-name: "/usr/local/rs/"))
 (set! $resource-root (make <root-dir> 
			    root-name: "[resource]/"
			    expanded-name: "/usr/local/rs/resource/")))

(let ((h (getenv "HOME")))
  (if h
      (set-expanded-name! $user-root (string-append h "/"))))

(let ((inst (rscheme-global-ref 29)))
  (set-expanded-name! $install-root (string-append inst "/"))
  (set-expanded-name! $resource-root (string-append inst "/resource/")))

(define-method pathname->string ((self <root-dir>))
    (root-name self))

;;

(define *special-roots* '())

(define (set-special-roots! lst)
  (set! *special-roots* lst))

(define (add-special-root! (entry <string>) (r <root-dir>))
  (set! *special-roots* (cons (cons entry r) *special-roots*)))

(%early-once-only
 (add-special-root! "[resource]" $resource-root)
 (add-special-root! "[install]" $install-root)
 (add-special-root! "~" $user-root)
 (add-special-root! "" $system-root))

(define (special-root (s <string>))
  (let ((a (assoc s *special-roots*)))
    (if a
	(cdr a)
	#f)))
