#|------------------------------------------------------------*-Scheme-*--|
 | File:    modules/repl/replcli.scm
 |
 |          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 |          as part of the RScheme project, licensed for free use
 |
 | Version: 1.21
 | Date:    1997.01.26 11:23:28
 | Build:   v0.7-0.9, 97.01.26
 |
 | Purpose: REPL command line interface
 `------------------------------------------------------------------------|#



;;(file-version-string "%Z%%W% 1.21 %G% 11:23:28")



(define-fluid *top-level-envt*)

;; there is a special protocol used by start/handler.scm,
;; where args == #f to suppress printing of the signal message
;; (because the handler has already printed it)

(define (repl-condition-handler (arg <condition>) next-handler)
  (display arg *console-output-port*)
  (let ((r (fluid-ref *cmd-loop* #f)))
    (if r
	(let ((breaks-deep (cmd-loop-depth r)))
	  (sub-loop r (if (eq? breaks-deep 0)
			  "break[~d]=>"
			  (string-append "break^"
					 (number->string (+ 1 breaks-deep))
					 "[~d]=>")))
	  ((cmd-loop-restart r)))
	(begin
	  (cmd-loop (fluid-ref *top-level-envt*) "break[~d]=>")
	  (process-exit 1)))))

(define (repl-interrupt)
  ;; we are called with RS signals disabled; re-enable them
  (os-set-sigenable #t)
  (let ((r (fluid-ref *cmd-loop* #f)))
    (if r
	(sub-loop r "interrupt[~d]=>")
	(cmd-loop (make-user-initial) "interrupt[~d]=>"))))

(define (restart-with path proc)
  (if (instance? path <file-name>)
      (restart-with (pathname->os-path path) proc)
      (begin
	(if (not (string? path))
	    (error "restart-with: path ~s not a string" path))
	(if (not (procedure? proc))
	    (error "restart-with: proc ~s not a procedure" proc))
	(let ((v (clone (rscheme-global-ref 0))))
	  (vector-set! v 4 proc)
	  (vector-set! (vector-ref v 0) 26 (rscheme-global-ref 26))
	  (generic-save-boot-image path v)
	  (values)))))

(define (generic-save-boot-image path vec)
  (let ((x (extension (string->file path))))
    (if (and x (string=? x "fas"))
	(if (module-loaded? 'fasl)
	    ((value (get-binding-from-module 'fasl 'save-fasl-image))
	     path
	     #f
	     (vector vec #f))
	    (error "~a: cannot save in fasl format" path))
	(save-image path vec '#() #f))))

(define *patch-stage?* #t)

(define *default-user-initial* #f)

(define (main args)
  (if (and *verbose* *license*)
      (repl-greeting))
  ;;
  ;; one-time
  ;;
  (if *patch-stage?*
      (begin
	(format #t "PATCHING...\n")
	;;
	;; patch up ourself
	;;
	(patch-symbols)
	;;
	;; construct some default modules, and set up for it
	;;
	(init-user-initial-modules)
	(set! *default-user-initial* (get-module 'user))
	;;
	;; don't do this next time
	;;
	(set! *patch-stage?* #f)))
  (if (and (pair? args)
	   (pair? (cdr args))
	   (string=? (car args) "--in"))
      (let ((m (cadr args)))
	;; note: we want to explicitly change the default user initial
	;; so that if an image is saved, when it is restarted it will
	;; be in the same environment
	(set! *default-user-initial*
	      (handler-case
	       (get-module (string->symbol m))
	       ((<condition> condition: c)
		(format $standard-error-port "** error trying to access module ~s\n" m)
		(display c $standard-error-port)
		*default-user-initial*)))
	(set! args (cddr args))))
  (run-repl args *default-user-initial*))

(define (run-repl args module)
  (fluid-let ((*console-input-port* (if *script*
					$console-input-port
					(open-edit-port)))
	      (*console-output-port* $console-output-port)
	      (*top-level-envt* (top-level-envt module)))
    (handler-bind (<condition> repl-condition-handler)
      (register-interrupt-handler! 'control-c repl-interrupt)
      (interpret-repl-args args (top-level-envt module)))))

;; made a part of the build-bootable process...
;;
;; (searches for a binding for main & sticks it in an rscheme-global)
;;
;;(set-main! repl-main)

(define (repl-greeting)
  (display "type \",warranty\" for details; type \",help\" for some help\n"))

(define (load-rc-files envt)
  (for-each (lambda (p)
	      (load-if-exists p envt))
	    (list (string->file "~/.R6.scm")
		  (string->file "[resource]/R6.scm"))))

(define (load-if-exists (path <file-name>) envt)
  (if (file-exists? path)
      (fluid-let ((*verbose* #f))  ;; [historical marker] this is the first
	                           ;; place where we used the new style of
                                   ;; fluid-let to it's advantage -- 
                                   ;; *verbose* was never a "fluid variable"
			           ;; but we can fluidly bind it anyway!
	(load-1 path envt))))

(define $default-prompt "top[~d]=>")

(define (run-script script envt args)
  (let* ((result (load-1 script envt))
	 (b (lookup envt 'main)))
    (if (and b (procedure? (value b)))
	((value b) args)
	result)))

;;
;;  this function processes the command line arguments to
;;  the REPL.  As its final action, it invokes a cmd-loop
;;  if one is to be run.
;;
;;  Among other things, it sets *args* the appropriate
;;  list of command-line arguments
;;

(define (interpret-repl-args args envt)
  (if (and *script*
	   (pair? args))
      ;;
      ;; if we're in script mode, all of the arguments get bound to *args*
      ;; and the rc file does NOT get loaded
      ;;
      (begin
	(set-app-args! (cdr args))
	(run-script (car args) envt (cdr args)))
      ;;
      ;; otherwise, interpret the arguments
      ;;
      (begin
	;;
	(load-if-exists
	 (string->file "[resource]/patch.scm") 
	 envt)
	;;
	;; load the ".rc" file, if it's not disabled
	;;
	(if (and (pair? args)
		 (string=? (car args) "-norc"))
	    (set! args (cdr args))
	    (load-rc-files envt))
	;;
	;; get the args following "--" first
	;;
	(let ((m (member "--" args)))
	  (if m
	      (set-app-args! (cdr m))))
	;;
	;; process the other command line arguments
	;;
	(let loop ((args args))
	  (if (null? args)
	      (cmd-loop envt $default-prompt)
	      (let (((a <string>) (car args)))
		(cond
		 ;;
		 ((string=? a "--")
		  ;;
		  ;; this has already been processed!, just go
		  ;; straight to the end
		  ;;
		  (loop '()))
		 ;;
		 ((string=? a "-exit")
		  (values))
		 ;;
		 ((string=? a "-e")
		  (eval-string (cadr args) envt)
		  (loop (cddr args)))
		 ;;
		 ;; -c[.repl] <path> [file...]
		 ;;
		 ;;  make a bootable image in <path>, with file... 
		 ;;  loaded.  bootable image will call main in
		 ;;  target image, or REPL main if -c.repl is specified
		 ;;
		 ;;  exits after generating the output file
		 ;;
		 ((or (string=? a "-c")
		      (string=? a "-c.repl"))
		  ;;
		  (for-each (lambda (a)
			      (if (and (> (string-length a) 1)
				       (eq? (string-ref a 0) #\+))
				  (use-in (string->symbol (substring a 1)) 
					  envt)
				  (load-1 a envt)))
			    (cddr args))
		  (warn-about-unbound-vars-created)
		  (format #t "saving to image: ~a\n" (cadr args))
		  (set-exceptions-abort-process! #f)
		  (restart-with (cadr args) (if (string=? a "-c.repl")
						main
						(eval-in-envt 'main envt)))
		  (values))
		 ;;
		 ((string=? a "-m")
		  (if (pair? (cdr args))
		      (begin
			(load-module-from-path (cadr args))
			(loop (cddr args)))
		      (error "missing argument for '-m' option")))
		 ;;
		 ((and (> (string-length a) 1)
		       (eq? (string-ref a 0) #\+))
		  (use-in (string->symbol (substring a 1)) envt)
		  (loop (cdr args)))
		 ;;
		 (else
		  (load-1 a envt)
		  (loop (cdr args))))))))))

(define (load-module-from-path (path <string>))
  (let ((i (string-search path #\=)))
    (bind ((module-name module-path (if i
					(values (string->symbol
						 (substring path 0 i))
						(substring path (+ i 1)))
					(values (string->symbol
						 (filename 
						  (string->file path)))
						path))))
      (link-load-module module-name (string->file module-path)))))
