#include "fasldef.h"

#ifdef PLATFORM_AIX

void init_stub_procs( struct FASL_Header *hdr, jump_addr stub )
{
  UINT_32 *f = (UINT_32 *)stub;
  hdr->stub_proc[0] = f[0];
  hdr->stub_proc[1] = f[1];
  hdr->stub_proc[2] = f[2];
}

#endif /* AIX */

/* linux, little-endian sunos boxes, and FreeBSD
 * run on intel hardware
 *  (for better abstraction, I should probably add an architecture
 *   configuration variable, like POWERPC, INTEL, ALPHA, etc.
 */

#if defined(PLATFORM_LINUX) \
     || (defined(PLATFORM_SUNOS) && defined(PLATFORM_IS_LITTLE_ENDIAN)) \
     || defined(__FreeBSD__)

void init_stub_procs( struct FASL_Header *hdr, jump_addr stub )
{
  static UINT_32 stub_ptr;
  UINT_8 *d = (UINT_8 *)hdr->stub_proc;
  UINT_32 addr;
  
  stub_ptr = (UINT_32)stub;
  addr = (UINT_32)&stub_ptr;

  *d++ = 0xff;  /* jmp absolute indirect (4-byte addr of fn ptr) */
  *d++ = 0x25;
  *d++ = addr & 0xFF;
  *d++ = (addr >> 8) & 0xFF;
  *d++ = (addr >> 16) & 0xFF;
  *d++ = (addr >> 24) & 0xFF;
}
#endif /* LINUX */

#if defined(PLATFORM_SUNOS) && !defined(PLATFORM_IS_LITTLE_ENDIAN)

/* SPARC 
            +-----+----------------------+
    call    | 0 1 |  29-bit displacement |
            +-----+----------------------+

stub gateway

    ==> save %sp, -112, %sp     0x9de3bf90
        call real_stub          <call>
        nop                     0x01000000
        ret                     0x81c7e008
        restore                 0x91e80008
*/

void init_stub_procs( struct FASL_Header *hdr, jump_addr stub )
{
  static UINT_32 stub_ptr;
  UINT_32 *d = (UINT_32 *)hdr->stub_proc;
  UINT_32 addr = (UINT_32)stub;

  *d++ = 0x9de3bf90;
  *d = ((addr - (UINT_32)d) / 4) + 0x40000000;
  d++;
  *d++ = 0x01000000;
  *d++ = 0x81c7e008;
  *d++ = 0x91e80008;
}

#endif /* SUNOS SPARC */
