/*------------------------------------------------------------------*-C-*-*
 * File:	packages/syscalls/syscallx.c
 * Version:	1.1
 * Date:	17:54:42
 * Build:	v-0.6-2.5, 97.01.26
 *
 * Purpose:	bytecode extensions to access generic system calls
 *------------------------------------------------------------------------*/

#ifdef INCL_SCCSID
char *F52RS_syscallx_sccsid__()
{ return "%W% 17:54:42"; }
#endif /* INCL_SCCSID */

#include <rscheme/scheme.h>
#include <rscheme/smemory.h>
#include <rscheme/bcextend.h>
#include <rscheme/osglue.h>

/*
   The Syscalls BCI Extension
*/

UINT_8 *bc_syscalls_extension( UINT_8 *pc, RS_bc_datum **args )
{
  RS_bc_datum *bot = *args;
  FILE *f;
  size_t n;

  switch (*pc++)
    {
    case 0: 
      bot->obj_val = FILE_AS_FX( stdin );
      bot++;
      break;
    case 1: 
      bot->obj_val = FILE_AS_FX( stdout );
      bot++;
      break;
    case 2: 
      bot->obj_val = FILE_AS_FX( stderr );
      bot++;
      break;
    case 3: /* fopen */
      bot -= 2;
      f = os_fopen( bot[0].raw_str_val, bot[1].raw_str_val );
      if (f)
	bot->obj_val = FILE_AS_FX(f);
      else
	bot->obj_val = FALSE_OBJ;
      bot++;
      break;
    case 4: /* fclose */
      fclose( FX_AS_FILE( bot[-1].obj_val ) );
      break;
    case 5: /* fwrite/str */
      bot -= 2;
      n = fwrite( string_text(bot[1].obj_val), 
		  1, string_length(bot[1].obj_val),
		  FX_AS_FILE(bot[0].obj_val) );
      bot[0].raw_int_val = n;
      bot++;
      break;
    case 6: /* fwrite/dec */
      bot -= 2;
      fprintf( FX_AS_FILE(bot[0].obj_val), 
	       "%d", 
	       fx2int( bot[1].obj_val ) );
      bot[0].raw_int_val = 0;
      bot++;
      break;
    case 7: /* fflush */
      bot[-1].raw_int_val = fflush( FX_AS_FILE( bot[-1].obj_val ) );
      break;
    case 8: /* ftell */
      bot[-1].raw_int_val = ftell( FX_AS_FILE( bot[-1].obj_val ) );
      break;
    case 9: /* fseek */
      bot -= 2;
      bot[-1].raw_int_val = fseek( FX_AS_FILE( bot[-1].obj_val ),
				   bot[0].raw_int_val,
				   bot[1].raw_int_val );
      break;
    case 10: /* fread-fill */
      bot -= 2;
      bot[-1].raw_int_val = fread( PTR_TO_DATAPTR( bot[0].obj_val ),
				   1, bot[1].raw_int_val,
				   FX_AS_FILE( bot[-1].obj_val ) );
      break;
    case 11: /* fputc */
      bot -= 2;
      fputc( GET_IMMEDIATE_VALUE(bot[1].obj_val),
	     FX_AS_FILE(bot[0].obj_val));
      break;
    case 12: /* fgetc */
      { int ch = fgetc( FX_AS_FILE(bot[-1].obj_val) );
	if (ch < 0)
	  bot[-1].obj_val = FALSE_OBJ;
	else
	  bot[-1].obj_val = MAKE_ASCII_CHAR(ch);
      }
      break;
    case 13: /* fgets */
      { char *s, temp[1024];
	s = fgets( temp, 1024, FX_AS_FILE(bot[-1].obj_val) );
	if (s)
	  bot[-1].obj_val = make_string(temp);
	else
	  bot[-1].obj_val = FALSE_OBJ;
      }
      break;
    case 14: /* popen */
      bot -= 2;
#ifdef HAS_POPEN
      f = (FILE *)popen( bot[0].raw_str_val, bot[1].raw_str_val );
      if (f)
	bot->obj_val = FILE_AS_FX(f);
      else
	bot->obj_val = FALSE_OBJ;
#else
      scheme_error( "popen: not implemented on this platform" );
#endif /* HAS_POPEN */
      bot++;
      break;
    case 15: /* fgetln */
      bot[-1].obj_val = fgetln( FX_AS_FILE(bot[-1].obj_val) );
      break;
    case 16: /* pclose */
#ifdef HAS_POPEN
      bot[-1].obj_val = int2fx( pclose( FX_AS_FILE( bot[-1].obj_val ) ) );
#else
      scheme_error( "pclose: not implemented on this platform" );
#endif
      break;
    }
  *args = bot;
  return pc;
}

#define QUICK_LEN  (1000)
#define EXTEND_LEN  (2000)

struct str_buf {
  char    *ptr, *limit, *base;
  UINT_32 length;
  obj     overflow;
  char    quick[QUICK_LEN];
};

static void str_buf_init( struct str_buf *b )
{
  b->ptr = b->base = b->quick;
  b->limit = b->base + QUICK_LEN;
  b->overflow = NIL_OBJ;
  b->length = 0;
}

static obj str_buf_flush( struct str_buf *b )
{
  UINT_32 N, len;
  obj result, over;
  char *dst;

  N = b->ptr - b->base;
  len = b->length + N;

  result = bvec_alloc( len+1, string_class );
  
  dst = ((char *)PTR_TO_DATAPTR(result)) + b->length;

  over = b->overflow;
  if (!EQ(over,NIL_OBJ))
    {
      do {
	memcpy( dst, (char *)PTR_TO_DATAPTR(pair_car(over)), N );

	over = pair_cdr(over);
	dst -= EXTEND_LEN;
	N = EXTEND_LEN;
      } while (!EQ(over,NIL_OBJ));
      N = QUICK_LEN;
      dst = ((char *)PTR_TO_DATAPTR(result));
    }
  memcpy( dst, b->quick, N );
  return result;
}

static void str_buf_addch( struct str_buf *b, char ch )
{
  if (b->ptr >= b->limit)
    {
      obj chunk;

      /* flush this buffer and add another */
      b->length += b->limit - b->base;
      chunk = bvec_alloc( EXTEND_LEN, byte_vector_class );
      b->base = b->ptr = (char *)PTR_TO_DATAPTR(chunk);
      b->limit = b->base + EXTEND_LEN;
      b->overflow = cons( chunk, b->overflow );
    }
  *(b->ptr)++ = ch;
}

obj fgetln( FILE *strm )
{
  struct str_buf B;
  int ch;

  ch = fgetc( strm );
  if (ch == EOF)
    {
      return FALSE_OBJ;
    }
  else if (ch == '\n')
    {
      return bvec_alloc( 1, string_class );
    }
  else
    {
      str_buf_init( &B );
      do {
	str_buf_addch( &B, ch );
	ch = fgetc( strm );
      } while (ch != EOF && ch != '\n');
      return str_buf_flush( &B );
    }
}
