
(define-unix-glue (dbm-open (path <raw-string>)
			    (mode <raw-int>)
			    (permissions <raw-int>))
{
DBM *db;

    db = dbm_open( path, rscheme_file_mode_to_os(mode), permissions );
    if (db)
	REG0 = RAW_PTR_TO_OBJ(db);
    else
	REG0 = FALSE_OBJ;
    RETURN1();
})

(define-unix-glue (dbm-fetch (db <DBM>) (key <string>))
{
datum d;

    d.dptr = string_text(key);
    d.dsize = string_length(key);
    d = dbm_fetch( db, d );
    if (d.dptr)
    {
	REG0 = bvec_alloc( d.dsize + 1, string_class );
	memcpy( PTR_TO_DATAPTR(REG0), d.dptr, d.dsize );
    }
    else
    {
	REG0 = FALSE_OBJ;
    }
    RETURN1();
})

(define-unix-glue (dbm-first-key (db <DBM>))
{
datum d;

    d = dbm_firstkey( db );
    if (d.dptr)
    {
	REG0 = bvec_alloc( d.dsize + 1, string_class );
	memcpy( PTR_TO_DATAPTR(REG0), d.dptr, d.dsize );
    }
    else
    {
	REG0 = FALSE_OBJ;
    }
    RETURN1();
})

(define-unix-glue (dbm-next-key (db <DBM>))
{
datum d;

    d = dbm_nextkey( db );
    if (d.dptr)
    {
	REG0 = bvec_alloc( d.dsize + 1, string_class );
	memcpy( PTR_TO_DATAPTR(REG0), d.dptr, d.dsize );
    }
    else
    {
	REG0 = FALSE_OBJ;
    }
    RETURN1();
})

(define-unix-glue (dbm-delete (db <DBM>) (key <string>))
{
datum d;
int rc;

    d.dptr = string_text(key);
    d.dsize = string_length(key);
    rc = dbm_delete( db, d );
    REG0 = rc ? FALSE_OBJ : TRUE_OBJ;
    RETURN1();
})

(define-unix-glue (dbm-store (db <DBM>) 
			     (key <string>) 
			     (value <string>)
			     insert_only_q)
    literals: ('key-present)
{
datum k, v;
int rc;

    k.dptr = string_text(key);
    k.dsize = string_length(key);
    v.dptr = string_text(value);
    v.dsize = string_length(value);
    rc = dbm_store( db, k, v,
		    truish(insert_only_q) ? DBM_INSERT : DBM_REPLACE );
    REG0 = (rc == 1) ? LITERAL(0) : (rc ? FALSE_OBJ : TRUE_OBJ);
    RETURN1();
})

(define-unix-glue (dbm-error (db <DBM>))
{
    REG0 = int2fx( dbm_error( db ) );
    RETURN1();
})

(define-unix-glue (dbm-close (db <DBM>))
{
    dbm_close( db );
    RETURN0();
})

(define-unix-glue (dbm-clear-error (db <DBM>))
{
    dbm_clearerr( db );
    REG0 = TRUE_OBJ;
    RETURN1();
})


;;  present a <table> interface

(define-class <DBM-table> (<table>)
    (DBM init-value: #f))

(define-syntax kwd 
  (syntax-form (k lst)
    (let ((m (memq (mquote k) lst)))
      (if m
	  (cadr m)
	  (error "required keyword `~s' not specified" (mquote k)))))
  (syntax-form (k lst dflt)
    (let ((m (memq (mquote k) lst)))
      (if m
	  (cadr m)
	  dflt))))

(define-method initialize ((self <DBM-table>) #rest iargs)
  (let ((path (kwd path: iargs))
	(flags (kwd flags: iargs '(read)))
	(mode (kwd mode: iargs #o666)))
    (set! flags (apply make-fd-open-mode flags))
    (let ((dbm (dbm-open path flags mode)))
      (if dbm
	  (set-DBM! self dbm)
	  (error "<DBM-table> initialize: ~a: ~a"
		 path (errorstr (errno))))
      self)))

(define-method table-insert! ((self <DBM-table>) 
			      (key <string>) 
			      (value <string>))
    (dbm-store (DBM self) key value #f))

(define-method table-lookup ((self <DBM-table>) (key <string>))
    (dbm-fetch (DBM self) key))

(define-method table-for-each ((self <DBM-table>) proc)
  (let ((dbm (DBM self)))
    (let loop ((key (dbm-first-key dbm)) (i 0))
	(if key
	    (begin
		(proc 0 ;; hash
			key
			(dbm-fetch dbm key))
		(loop (dbm-next-key dbm) (+ i 1)))
	    i))))

(define-method table-lookup ((self <DBM-table>) (key <string>))
    (dbm-fetch (DBM self) key))

(define-method key-sequence ((self <DBM-table>))
  (let ((dbm (DBM self)))
    (let loop ((key (dbm-first-key dbm)) (r '()))
	(if key
	    (loop (dbm-next-key dbm) (cons key r))
	    r))))
  
