/*------------------------------------------------------------------*-C-*-*
 * File:    handc/dfltmain/findimag.c
 *
 *          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 *          as part of the RScheme project, licensed for free use
 *
 * Version: 1.3
 * Date:    1996.07.25 21:27:07
 * Build:   v0.7-0.9, 97.01.26
 *
 * Purpose: function to locate system image
 *------------------------------------------------------------------------*/




#ifdef INCL_SCCSID
char *Fb1RS_dist_dfltmain_main_sccsid__()
{ return "%Z%%W% 1.3 %G% 21:27:07"; }
#endif /* INCL_SCCSID */

#include <rscheme/osglue.h>
#include <rscheme/rlseconf.h>
#include <rscheme/scheme.h>
#include <rscheme/api.h>

#include <string.h>
#include <ctype.h>
#ifdef __NeXT__
#include <mach/mach.h>
#endif

const char *find_system_image( const char *arg0, const char *dflt );

/* try to locate the system image if it
   wasn't specified on the command line
*/

const char *find_system_image( const char *arg0, const char *dflt )
{
static char temp[1024];
char *p;
#ifdef __NeXT__
int n;

   if (getsectdata( "IMAGE", "system", &n ))
     {
       return "sect:IMAGE:system";
     }
#endif

  /* if the default is specified and exists, use it
   *
   * this is used by applications that supply their own
   * image
   */

  if (dflt && os_file_exists_p(dflt))
     return dflt;

  /* look in ~/lib/rs/$RSCHEME_VERSION/system.img  */

  if (os_getenv("HOME"))
    {
      strcpy( temp, os_getenv( "HOME" ) );
      strcat( temp, "/lib/rs/" RSCHEME_VERSION "/system.img" );
      if (os_file_exists_p(temp))
	return temp;
    }

  /* look in [install]/resource/system.img */

  strcpy( temp, rs_install_dir );
  strcat( temp, "/resource/system.img" );
  if (os_file_exists_p(temp))
    return temp;

  /*  If argv[0] is of the form foo/bin/blech
      then look for foo/resource/blech.img
  */

  if ((p = strrchr(arg0,'/')))
    {
      char *p2;

      strncpy( temp, arg0, p - arg0 );
      temp[p - arg0] = 0;

      p2 = strrchr( temp, '/' );
      if (p2)
	p2[1] = 0;

      strcat( temp, "resource/" );
      strcat( temp, p+1 );
      strcat( temp, ".img" );
    }
  else
    strcpy( temp, "../resource/system.img" );

  if (os_file_exists_p(temp))
    return temp;

  return NULL;
}
