/*------------------------------------------------------------------*-C-*-*
 * File:    handc/heapi/saveq.c
 *
 *          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 *          as part of the RScheme project, licensed for free use
 *
 * Version: 1.3
 * Date:    1996.04.18 20:39:58
 * Build:   v0.7-0.9, 97.01.26
 *
 * Purpose: queue of objects being saved
 *------------------------------------------------------------------------*/




#ifdef INCL_SCCSID
char *F41RS_dist_heapi_saveq_sccsid__()
{ return "%Z%%W% 1.3 %G% 20:39:58"; }
#endif /* INCL_SCCSID */

#include <stdio.h>
#include <stdlib.h>
#include "saveq.h"

/* the -4 is because in struct SaveQueue_item is char data[4] */

#define EXTRA_SIZE	(sizeof( struct SaveQueue_item ) - 4)

void init_queue( SaveQueue *q, unsigned item_size, ncanonicalize_fn *fn )
{
    q->front = NULL;
    q->tail = NULL;
    q->item_size = item_size + EXTRA_SIZE;
    q->netcanon_fn = fn;
}

void *enqueue_item( SaveQueue *q )
{
struct SaveQueue_item *x;

    x = (struct SaveQueue_item *)malloc( q->item_size );
    if (q->tail)
	q->tail->next = x;
    else
	q->front = x;
    q->tail = x;
    x->next = NULL;
    return x->data;
}


void init_queue_state( SaveQueueIterator *state, SaveQueue *q )
{
    state->q = q;
    state->p = NULL;
}

const void *dequeue_item( SaveQueueIterator *state )
{
    if (!state->p)
    {
	state->p = state->q->front;
    }
    else
    {
	state->p = state->p->next;
    }
    return (state->p) ? state->p->data : NULL;
}

unsigned fwrite_queue_contents( SaveQueue *q, FILE *f )
{
unsigned n, per_item;
struct SaveQueue_item *i;
#ifndef PLATFORM_IS_BIG_ENDIAN
ncanonicalize_fn *canonicalize = q->netcanon_fn;
#endif

    n = 0;
    per_item = q->item_size - EXTRA_SIZE;
    for (i=q->front; i; i=i->next)
    {
#ifdef PLATFORM_IS_BIG_ENDIAN
	fwrite( i->data, 1, per_item, f );
#else
	fwrite( canonicalize( i->data ), 1, per_item, f );
#endif
	n++;
    }
    return n;    
}

void free_queue( SaveQueue *q )
{
struct SaveQueue_item *i, *j;

    for (i=q->front; i; i=j)
    {
	j = i->next;
	free( i );
    }
}

#if 0
void test( const char *teststr )
{
SaveQueue q;
SaveQueueIterator i;
int *p;

    init_queue( &q, sizeof(int) );
    init_queue_state( &i, &q );
    
    printf( "From '%s': ", teststr );
    while (*teststr)
    {
        if (*teststr == '.')
	{
	    p = dequeue_item(&i);
	    if (p)
		printf( " (%d)", *p );
	    else
		printf( " ()" );
	}
	else
	{
	    p = enqueue_item( &q );
	    *p = *teststr - '0';
	    printf( " %d", *p );
	}
	teststr++;
    }
    printf( "\nAgain: " );
    init_queue_state( &i, &q );
    while (p = dequeue_item(&i))
    {
	printf( " %d", *p );
    }
    printf( "\n" );
}

int main( int argc, const char **argv )
{
int i;

    for (i=1; i<argc; i++)
	test( argv[i] );
    return 0;
}
#endif
