-- caveats
-- - platform config is partially hard-coded; see setupPlatform

-- uses these Scripting Additions:
-- - Finder Scripting Extension
-- - Display Dialog
-- - Choose Files and Folders
-- - File Commands

-- AppleScript separates underlines from alphanumerics, so it parses
-- lists of unix (and Mac!) filenames incorrectly. There may be some
-- incantation rendering this unecessary, but I don't know it.
on unixWords(str)
	local w, i, j, len
	set len to the length of str
	set w to {}
	set i to 1
	set j to 1
	repeat while j < len
		if (character j of str is in {" ", tab}) then
			if i < j then
				set w to {((characters i thru (j - 1) of str) as string)} & w
			end if
			set i to j + 1
		end if
		set j to j + 1
	end repeat
	if i < j then
		set w to {((characters i thru j of str) as string)} & w
	end if
	return w
end unixWords

on getMakefileCFILES(fileSpec)
	try
		set fd to open for access fileSpec as alias
		set allProcessed to false
		set processingList to false
		set projectFiles to ""
		repeat until allProcessed
			set nextLine to read fd as string before return
			if nextLine contains "CFILES=" then
				set processingList to true
				set nextLine to (characters 8 thru (length of nextLine) of nextLine) as string
			end if
			if processingList and the last character of nextLine is not "\\" then
				set allProcessed to true
			else if processingList then
				set nextLine to (characters 1 thru ((length of nextLine) - 1) of nextLine) as string
			end if
			if processingList then
				set projectFiles to projectFiles & " " & nextLine
			end if
		end repeat
		close access fd
	on error errText
		log "fuck! " & errText
	end try
	set fileList to unixWords(projectFiles)
	return fileList
end getMakefileCFILES

on linkItem(fileName, srcItem, destFolder)
	tell application "Finder"
		if exists (a reference to the item named fileName of destFolder) then
			set oldFile to get the item named fileName of destFolder
			if the class of oldFile is not alias file then
				error "cannot replace file with alias"
			end if
			set oldFile to the original item of oldFile
			if not (oldFile = srcItem) then
				delete oldFile
			end if
		end if
		if not (exists (a reference to the item named fileName of destFolder)) then
			make new alias at destFolder to srcItem
			set the name of the result to fileName
			log "linked file " & fileName
		else
			log "left file " & fileName & " alone: already correctly linked"
		end if
	end tell
end linkItem

on setupPlatform(projectFolder, platform)
	-- create links from config-dependent source dirs to main source tree
	tell application "Finder"
		set configFolder to the folder named "cfg" of projectFolder
		set gcFolder to the folder named "irc" of the folder named "gc" of configFolder
		set gcadaptFolder to the folder named "irc" of the folder named "gcadapt" of configFolder
		set rdlnFolder to the folder named "std" of the folder named "rdln" of configFolder
	end tell
	linkItem("gc", gcFolder, projectFolder)
	linkItem("gcadapt", gcadaptFolder, projectFolder)
	linkItem("rdln", rdlnFolder, projectFolder)
	
	-- create links from platform-specific files to source dirs
	tell application "Finder"
		set platformDir to the folder named platform of the folder named "platform" of projectFolder
		set osglue to the file named "osglue.c" of platformDir
		set mapf to the file named "mapf.c" of platformDir
		set platform to the file named "platform.h" of platformDir
		set macHeaders to the file named "rs-mac-headers.h" of platformDir
		set rlseconf to the file named "rlseconf.h" of platformDir
		set buildsty to the file named "buildsty.h" of platformDir
		set modulo to the file named "modulo.h" of platformDir
		set runtime to the folder named "runtime" of projectFolder
		set heapi to the folder named "heapi" of projectFolder
		set installFolder to the folder named "install" of projectFolder
		set includeFolder to the folder named "include" of installFolder
		set rscheme to the folder named "rscheme" of includeFolder
		set stdModule to the file named "stdmodul.h" of rscheme
	end tell
	linkItem("osglue.c", osglue, runtime)
	linkItem("mapf.c", mapf, heapi)
	linkItem("rs-mac-headers.h", macHeaders, includeFolder)
	linkItem("rscheme/buildsty.h", buildsty, rscheme)
	linkItem("rscheme/modulo.h", modulo, rscheme)
	linkItem("rscheme/platform.h", platform, rscheme)
	linkItem("rscheme/rlseconf.h", rlseconf, rscheme)
	linkItem("rscheme/stdmodul.h", stdModule, rscheme)
end setupPlatform

script doItemList
	property itemListFile : false
	
	on setup(itemList)
		set itemListFile to itemList
	end setup
	
	on doOne(thing)
		log "override me" & thing
	end doOne
	
	if itemListFile = false then
		log "wasn't able to access file"
	else
		set fd to open for access the itemListFile as alias
		set itemList to read fd as string using delimiters {" ", return}
		close access fd
		repeat with thing in itemList
			doOne(thing)
		end repeat
	end if
end script -- doItemList

script doSubProjects
	property parent : doItemList
	property projectFolder : false
	
	on setup(proj)
		set projectFolder to proj
		tell application "Finder" to get the file "subproj.lst" of proj
		continue setup(the result)
	end setup
	
	on doOne(subprojName)
		tell application "Finder" to get the item named subprojName of projectFolder
		log the result
	end doOne
	
	continue run {}
end script -- doSubProjects

script installSubprojHeaders
	property parent : doItemList
	property subprojFolder : false
	property destFolder : false
	
	on setup(proj, dest)
		local includesFile
		set subprojFolder to proj
		set destFolder to dest
		set includesFile to false
		try
			tell application "Finder" to get the file "includes.lst" of proj
			set includesFile to the result
		on error errMsg
			log errMsg
			set subprojFolder to false
			set destFolder to false
		end try
		continue setup(includesFile)
	end setup
	
	on doOne(headerName)
		if subprojFolder = false then
			log "includes.lst didn't exist"
		else
			try
				tell application "Finder" to get the file named headerName of subprojFolder
				linkItem("rscheme/" & headerName, the result, destFolder)
			on error errMsg
				log "couldn't link!" & headerName & errMsg
			end try
		end if
	end doOne
	
	continue run {}
end script -- installSubprojHeaders

script installHeaders
	property parent : doSubProjects
	property projectFolder : false
	property installFolder : false
	
	on setup(proj)
		set projectFolder to proj
		tell application "Finder" to set installFolder to the folder "rscheme" of the folder "include" of the folder "install" of projectFolder
		continue setup(proj)
	end setup
	
	on doOne(subprojName)
		tell application "Finder"
			get the item named subprojName of projectFolder
			set subprojFolder to the result
			if the class of subprojFolder is alias file then
				set subprojFolder to the original item of subprojFolder
			end if
		end tell
		tell installSubprojHeaders to setup(subprojFolder, installFolder)
		tell installSubprojHeaders to run
	end doOne
	
	continue run {}
end script -- installHeaders

on createProject(srcFolder, platform, targetName)
	
	-- find a stationery project to clone
	set useStationery to true
	get the path to the application "CodeWarrior IDE 1.5b2" as alias
	tell application "Finder" to get the folder of the result
	set CWFolder to the result as alias
	set default folder CWFolder -- this doesn't seem to be working
	try
		set projectStationery to choose file with prompt "Choose a stationery project:"
	on error errText number errNum
		set useStationery to false
	end try
	
	-- create a new project in :install:lib
	display dialog "Project Name:" default answer ("rslib-" & targetName & ".") 
		buttons ["Cancel", "OK"] default button "OK"
	set newProjectReply to the result
	if button returned of newProjectReply is "Cancel" then
		error "Project Creation aborted"
	end if
	-- create :install:lib if necessary
	tell application "Finder"
		get (the folder "install" of srcFolder) as reference
		set installFolder to the result
		log installFolder
		if exists (a reference to the folder "lib" of installFolder) then
			get (the folder "lib" of installFolder) as reference
			set libFolder to the result
		else
			make new folder at installFolder 
				with properties {name:"lib"}
			set libFolder to the result
		end if
		log libFolder
	end tell
	set projectPath to (libFolder as string) & (text returned of newProjectReply)
	log projectPath
	-- close any project now open
	try
		tell application "CodeWarrior IDE 1.5b2" to 
			Close Project
	on error errText
		log errText
	end try
	-- now create the project
	if useStationery then
		tell application "CodeWarrior IDE 1.5b2" to 
			Create Project projectPath from stationery (projectStationery as alias)
	else
		tell application "CodeWarrior IDE 1.5b2" to Create Project projectPath
	end if
	
	-- this is probably unecessary
	tell application "CodeWarrior IDE 1.5b2" to Reset File Paths
	
	-- set up the access paths for our project
	tell application "CodeWarrior IDE 1.5b2"
		-- gotta get both here, even though we're zeroing out User Paths
		Get Preferences of {User Paths, System Paths} from panel "Access Paths"
		set paths to the result
		set User Paths of paths to {}
		set System Paths of paths to 
			{{name:"::include", recursive:true, origin:project relative}} 
				& System Paths of paths
		Set Preferences of panel "Access Paths" to paths
	end tell
	
	-- specify the library name, if self-evident
	tell application "CodeWarrior IDE 1.5b2" to set f to Get Project Specifier
	tell application "Finder" to set fname to the name of f
	if fname ends with "." then
		set pname to (characters 1 thru ((length of fname) - 2) of fname) as string
		tell application "CodeWarrior IDE 1.5b2" to 
			Set Preferences of panel 
				(targetName & " Project") to {File Name:pname}
	else
		log "couldn't create target file name from " & fname
	end if
	
	-- remove any Sources segment
	try
		tell application "CodeWarrior IDE 1.5b2" to 
			Remove Files ["<replace me ANSI>.c"]
	on error errText number errNum
		log errText
	end try
	try
		tell application "CodeWarrior IDE 1.5b2" to 
			Remove Files ["<replace me Mac>.c"]
	on error errText number errNum
		log errText
	end try
	
	-- set up prefix file
	tell application "CodeWarrior IDE 1.5b2"
		Set Preferences of panel "C/C++ Compiler" to {Prefix File:"rs-mac-headers.h"}
	end tell
	
	-- set project type
	tell application "CodeWarrior IDE 1.5b2"
		Set Preferences of panel (targetName & " Project") to {Project Type:library}
	end tell
	
end createProject

script addSubprojSources
	property itemListFile : false
	property segmentNumber : false
	property segmentName : false
	property sourceFolder : false
	
	on setup(subproj, segNum, segName)
		try
			tell application "Finder" to get the file "Makefile" of subproj
		on error errText number errNum
			try
				tell application "Finder" to get the file "Makefile.in" of subproj
			on error errText number errNum
				error errText number errNum
			end try
		end try
		set itemListFile to the result
		set sourceFolder to subproj
		set segmentNumber to segNum
		set segmentName to segName
	end setup
	
	set theFiles to getMakefileCFILES(itemListFile)
	set fileAliases to {}
	repeat with f in theFiles
		tell application "Finder" to get the file named f of sourceFolder
		set fileAliases to fileAliases & {the result as alias}
	end repeat
	log fileAliases
	tell application "CodeWarrior IDE 1.5b2"
		-- here we rely on the undocumented behavior of CW that
		-- files are added to a new segment; it could just as easily
		-- be (in some future CW version) that files are added to 
		-- some random existing segment, in which case this code will break.
		Add Files fileAliases
		Set Segment segmentNumber to {class:Segment, name:segmentName}
	end tell
	
end script -- addSubprojSources

script doSources
	property parent : doSubProjects
	property projectFolder : false
	property segmentNumber : 1
	
	on setup(proj)
		set projectFolder to proj
		tell application "CodeWarrior IDE 1.5b2" to Get Segments
		set segmentNumber to (length of the result) + 1
		continue setup(proj)
	end setup
	
	on doOne(subprojName)
		tell application "Finder"
			get the item named subprojName of projectFolder
			set subprojFolder to the result
			if the class of subprojFolder is alias file then
				set subprojFolder to the original item of subprojFolder
			end if
		end tell
		tell addSubprojSources to setup(subprojFolder, segmentNumber, subprojName)
		tell addSubprojSources to run
		set segmentNumber to segmentNumber + 1
	end doOne
	
	continue run {}
end script -- doSources

on configurePlatform(srcFolder, platformName)
	setupPlatform(srcFolder, platformName)
	tell installHeaders to setup(srcFolder)
	tell installHeaders to run
end configurePlatform

on configureProject(srcFolder, platform, target)
	createProject(srcFolder, platform, target)
	tell doSources to setup(srcFolder)
	tell doSources to run
end configureProject

property projectFolder : false
--tell application "Finder" to get the folder named "tule:Development:rs-0.6-1.4:dist"
if projectFolder is not false then
	set default folder projectFolder
end if
choose folder with prompt "Please choose the RScheme folder you would like to operate on"
set projectFolder to the result

display dialog "Configure Macintosh platform?" buttons ["Yes", "No"] default button "Yes"
if the button returned of the result is "Yes" then
	configurePlatform(projectFolder, "mac")
end if

display dialog "Create CW Project?" buttons ["Yes", "No"] default button "Yes"
if the button returned of the result is "Yes" then
	display dialog "Which Target?" buttons ["68k", "PPC"] default button "68k"
	configureProject(projectFolder, "mac", the button returned of the result)
	display dialog "Build CW Project?" buttons ["Yes", "No"] default button "Yes"
	if the button returned of the result is "Yes" then
		with timeout of 100000 seconds
			tell application "CodeWarrior IDE 1.5b2" to Make Project
		end timeout
	end if
end if
