/*------------------------------------------------------------------*-C-*-*
 * File:    handc/platform/mac/osglue.c
 *
 *          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 *          as part of the RScheme project, licensed for free use
 *
 * Version: 1.11
 * Date:    1996.04.18 20:47:06
 * Build:   v0.7-0.9, 97.01.26
 *
 * Purpose: Low-level OS glue for Macintosh
 *------------------------------------------------------------------------*/




#ifdef INCL_SCCSID
char *F24RS_dist_platform_mac_osglue_sccsid__()
{ return "%Z%%W% 1.11 %G% 20:47:06"; }
#endif /* INCL_SCCSID */

#include "scheme.h"
#include "intrs.h"
#include "regs.h"

#include "runtime.h"
#include "osglue.h"

#include <Timer.h>
#include <stdlib.h>
#include <string.h>
#include <unix.h>

#define MAX_SIG (10)

rs_bool rssig_ready = NO;

struct sig_queue {
  rs_bool *ready;
  rs_bool enable;
  unsigned num_sig;
  int sig_queue[MAX_SIG];
};

static void sig_add( struct sig_queue *q, int id );

struct timer_info {
  struct TMTask hdr;
  struct sig_queue *q;
};

static struct sig_queue the_sig_queue;
static rs_bool timer_task_installed_p = NO;
static struct timer_info the_timer;


static void osglue_error( const char *fn, int rc )
{
    scheme_error( "OS glue error in ~s: rc ~d",
    		  2, make_string(fn), int2fx(rc) );
}


const char *os_getenv( const char *key )
{
    return NULL;
}

static char *make_macpath( char *temp, const char *path )
{
  if (strncmp( path, "./", 2 ) == 0)
    path += 2;
	
  if (strchr( path, '/' ))
    {
      char *p = temp;
      
      if (path[0] == '/')
	{
	  path++;
	}
      else
	{
	  *p++ = ':';
	}
      while (*path)
	{
	  if (*path == '/')
	    *p++ = ':';
	  else
	    *p++ = *path;
	  path++;
	}
      *p++ = 0;
    }
  else
    {
      strcpy( temp, path );
    }
  return temp;
}

/* check to see if a file exists */

rs_bool os_file_exists_p( const char *path )
{
  FILE *f;
  char temp[1000];

  f = fopen( make_macpath( temp, path ), "r" );
  /* printf( "[stat '%s']", temp ); */
	
  if (f)
    {
      fclose(f);
      return YES;
    }
  else
    return NO;
}

/* convert a unix-standard pathname to local OS format */

obj os_path( obj path )
{
  char temp[1000];
  return make_string( make_macpath( temp, string_text(path) ) );
}


/* open a file using unix-std pathnames */

FILE *os_fopen( const char *path, const char *mode )
{
  char temp[1000];
  return fopen( make_macpath( temp, path ), mode );
}


/* the real-time timer... */

UINT_32 system_timeout;

UINT_32 os_sleep( UINT_32 msec )
{
  unsigned t = (msec + 500) / 1000;
  sleep( t );
  return t * 1000;
}

static void exit_handler( void )
{
  os_halt_timer();
}

static void timeout_handler( void *self )
{
  sig_add( ((struct timer_info *)self)->q, RSSIG_TIMEOUT );
}

static void init_timer_task( void )
{
  the_timer.hdr.qLink = NULL;
  the_timer.hdr.qType = 0;
  the_timer.hdr.tmCount = 0;
  the_timer.hdr.tmWakeUp = 0;
  the_timer.hdr.tmReserved = 0;
  the_timer.hdr.tmAddr = NewTimerProc(timeout_handler);
  the_timer.q = &the_sig_queue;
}

void init_os( void )
{
  the_sig_queue.num_sig = 0;
  the_sig_queue.enable = YES;
  the_sig_queue.ready = &rssig_ready;

  atexit( &exit_handler );
  init_timer_task();
}


static obj cis( const char *str, obj rest )
{
  return cons( intern( make_string(str) ), rest );
}

obj os_type( void )
{
  return cis( "mac", NIL_OBJ );
}

obj os_getwd( void )
{
  char *p, temp[1000];
  
  getcwd( temp+1, 1000 );

  /* convert to unixy format */

  p = temp;
  *p++ = '/';
  while (*p)
    {
      if (*p == ':')
	*p = '/';
      p++;
    }

  /* strip off trailing '/' -- we know its a directory */

  if (p[-1] == '/')
    p[-1] = 0;

  return make_string( temp );
}

/* change the current working directory,
   or signal an error on failure 
*/

void os_setwd( const char *path )
{
  char temp[1000];

  if (chdir( make_macpath(temp, path) ) < 0)
    scheme_error( "os_setwd(~s) failed", 1, make_string(path) );
}



#define BEGIN_CRITICAL() do { /* what to do..? */
#define END_CRITICAL()  } while (0)

static obj sig_flush( struct sig_queue *q )
{
  int siginfo;
  obj lst = NIL_OBJ;

  if (q->num_sig)
    {
      unsigned n;

      BEGIN_CRITICAL();
      n = q->num_sig;
      if (n > MAX_SIG)
	{
	  lst = cons( lookup_symbol( "signal-queue-overflow" ), lst );
	  n = MAX_SIG;
	}
      while (n > 0)
	{
	  lst = cons( int2fx( q->sig_queue[--n] ), lst );
	}
      q->num_sig = 0;
      *(q->ready) = NO;
      END_CRITICAL();
    }
  return lst;
}


obj os_flush_sig_queue( void )
{
  return sig_flush( &the_sig_queue );
}


static void sig_add( struct sig_queue *q, int id )
{
  BEGIN_CRITICAL();
  if (q->num_sig >= MAX_SIG)
    {
      q->sig_queue[q->num_sig] = id;
    }
  else
    {
      fprintf( stderr, "\ninterrupts not being handled in a timely manner\n" );
      fprintf( stderr, "** aborting scheme process **\n" );
      abort();
    }
  q->num_sig++;
  if (q->enable)
    *q->ready = YES;
  END_CRITICAL();
}

void os_enqueue_sig( int id )
{
  sig_add( &the_sig_queue, id );
}


rs_bool os_set_sigenable( rs_bool flag )
{
  rs_bool old;

  BEGIN_CRITICAL();

  old = the_sig_queue.enable;

  the_sig_queue.enable = flag;
  if (flag)
    {
      if (the_sig_queue.num_sig)
	*the_sig_queue.ready = YES;
    }
  else
    {
      *the_sig_queue.ready = NO;
    }
  END_CRITICAL();
  return old;
}

UINT_32 os_halt_timer( void )
{
  UINT_32 ms;

  if (timer_task_installed_p)
    {
      /* remove the timer task */
      timer_task_installed_p = NO;
      RmvTime( (QElemPtr)&the_timer );
    }
  else
    ms = 0;

  /* if tmCount is negative,
   * then it's just in units of microseconds 
   */

  ms = (the_timer.hdr.tmCount < 0) ?
    -(the_timer.hdr.tmCount/1000) :
      the_timer.hdr.tmCount;
  
  if (ms == 0)
      {
	unsigned i;
	rs_bool qd = NO;

	/* make sure it's been queued 
	 * (note: because we've just turned off the timer, we
	 * don't have to worry about missing the insertion
	 * of the sig after we start looking, e.g., if the compiler
	 * caches num_sig)  (Also, we know that os_flush_sig_queue()
	 * won't be called at interrupt level)
	 */

	for (i=0; i<the_sig_queue.num_sig; i++)
	  if (the_sig_queue.sig_queue[i] == RSSIG_TIMEOUT)
	    {
	      qd = YES;
	      break;
	    }
	if (!qd)
	  os_enqueue_sig( RSSIG_TIMEOUT );
      }
   return ms;
}

void os_set_timer( UINT_32 msec )
{
  if (timer_task_installed_p)
    {
      RmvTime( (QElemPtr)&the_timer );
    }
  timer_task_installed_p = YES;
  InsTime((QElemPtr) &the_timer);

  PrimeTime((QElemPtr) &the_timer, msec);
}

UINT_32 os_get_time_remaining( void )
{
  long int remaining_time;
  
  if (timer_task_installed_p)
    {
      UINT_32 ms = os_halt_timer();
      os_set_timer(ms);
      return ms;
    }
  else
    return 0;
}
