/*-----------------------------------------------------------------*-C-*---
 * File:    handc/runtime/classes.c
 *
 *          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 *          as part of the RScheme project, licensed for free use
 *
 * Version: 1.1
 * Date:    1996.05.24 12:49:09
 * Build:   v0.7-0.9, 97.01.26
 *
 * Purpose: Runtime system support for <<class>> objects
 *------------------------------------------------------------------------*/

#include <rscheme/scheme.h>
#include <rscheme/smemory.h>

/* helper function... called by `object_class' when the
   given instance does not have a PTR tag
*/

obj immob_class( obj thing )
{
  if (OBJ_ISA_FIXNUM(thing))
    return fixnum_class;
  else if (OBJ_ISA_IMMOB(thing))
    {
      switch (SECONDARY_TAG(thing))
	{
	case BOOLEAN_TAG:		return boolean_class;
	case NIL_TAG:		return nil_class;
	case ASCII_CHAR_TAG:	return ascii_char_class;
	case UNICODE_CHAR_TAG:	return unicode_char_class;
	case UNIQUE_OBJ_TAG:	return unique_obj_class;
	  
	case SPARE_1_TAG:	return spare_1_class;
	case SPARE_2_TAG:	return spare_2_class;
	case SPARE_3_TAG:	return spare_3_class;
	}
    }
  return unique_obj_class;
}

rs_bool CLASS_P( obj thing )
{
obj class1;

  if (!OBJ_ISA_PTR(thing))
    return NO;
  class1 = CLASSOF_PTR(thing);

  /* allow only one level of subclassing <<class>> */

  if (EQ(class1, class_class))
    return YES;
  class1 = gvec_read(class1,SLOT(3));
  if (!PAIR_P(class1))
    return NO;
  return EQ(pair_car(class1),class_class);
}

unsigned class_image_mode( obj a_class )
{
    return fx2int( gvec_read( a_class, SLOT(2) ) );
}

rs_bool class_is_gvec( obj a_class )
{
    assert( CLASS_P(a_class) );
    return EQ(gvec_read(a_class,SLOT(1)),ZERO);
}

obj class_name( obj a_class )
{
    assert( CLASS_P(a_class) );
    return gvec_read( a_class, SLOT(0) );
}

/* returns YES iff class1 is a SUBCLASS of class2
   (improper subclassing *NOT* permitted, having already
    been checked by subclass_p).  Also, this function
   will crash if class_supers() returns something besides
   a <list>
*/

rs_bool indirect_subclass_p( obj class1, obj class2 )
{
  obj sup;
  
  while (1)
    {
      sup = class_supers(class1);
      if (NULL_P(sup))
	return NO;

      class1 = pair_car( sup );
      if (EQ(class1, class2))
	return YES;

      /* unrolled by hand once... */

      sup = class_supers(class1);
      if (NULL_P(sup))
	return NO;

      class1 = pair_car( sup );
      if (EQ(class1, class2))
	return YES;
    }
}


