/*------------------------------------------------------------------*-C-*-*
 * File:    handc/runtime/intrs.c
 *
 *          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 *          as part of the RScheme project, licensed for free use
 *
 * Version: 1.13
 * Date:    1996.12.20 15:52:33
 * Build:   v0.7-0.9, 97.01.26
 *
 * Purpose: Interrupt handling interface
 *------------------------------------------------------------------------*/

#ifdef INCL_SCCSID
char *F14RS_dist_runtime_intrs_sccsid__()
{ return "%Z%%W% 1.13 %G% 15:52:33"; }
#endif /* INCL_SCCSID */

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <rscheme/intrs.h>
#include <rscheme/smemory.h>
#include <rscheme/vinsns.h>
#include <rscheme/runtime.h>
#include <rscheme/scheme.h>
#include <rscheme/osglue.h>
#if HAVE_WAIT
#include <sys/types.h>
#include <sys/wait.h>
#endif

#ifdef SIGUSR_HOOKS
int sigusr1_flag, sigusr2_flag;
#endif /* SIGUSR_HOOKS */

#ifdef __cplusplus
extern "C" {
#endif

/* static void set_interrupt_flag( int sig, int code, 
   struct sigcontext *scp, 
   char *addr )*/

static void set_interrupt_flag( int x )	/* signal handler */
{
  struct RSSIG_info sig;
  sig.signum = RSSIG_USER_INTR;
  os_enqueue_sig_from_handler( &sig );
}

#if HAVE_WAIT
#ifdef PLATFORM_NEXT
typedef int pid_t;
#endif

static void capture_child_exit_status( int x )
{
  struct RSSIG_info sig;
  pid_t proc;
  int status;

  while (1)
    {
      /* drain the queue of waiting children */
      proc = waitpid( -1, &status, WNOHANG );
      if (proc <= 0)
	break;

      sig.signum = RSSIG_CHILD_EXITED;
      sig.data.child_exited.process_id = proc;
      sig.data.child_exited.status = status;
      
      os_enqueue_sig_from_handler( &sig );
    }
}
#endif

#ifdef SIGUSR_HOOKS
static void set_sigusr_flag( int x )	/* signal handler */
{
  if (x == SIGUSR1)
    sigusr1_flag = 1;
  else if (x == SIGUSR2)
    sigusr2_flag = 1;
}
#endif /* SIGUSR_HOOKS */

/* used by corelib/intrglue.scm//setup-signal-handler */

void c_signal_catcher( int signum )
{
  struct RSSIG_info sig;

  sig.signum = RSSIG_C_SIGNAL;
  sig.data.c_signal.signal_number = signum;
  
  os_enqueue_sig_from_handler( &sig );
}

#ifdef __cplusplus
}
#endif

void init_interrupts( void )
{
  os_register_signal_handler( SIGINT, set_interrupt_flag );
#ifdef SIGUSR_HOOKS
  os_register_signal_handler( SIGUSR1, set_sigusr_flag );
  os_register_signal_handler( SIGUSR2, set_sigusr_flag );
#endif /* SIGUSR_HOOKS */
}

void enable_subprocess_capture( void )
{
#if HAVE_WAIT
  os_register_signal_handler( SIGCHLD, capture_child_exit_status );
#endif
}

jump_addr dispatch_interrupt( jump_addr would_be_next )
{
jump_addr f;
struct RSSIG_info nxt;
unsigned n = 0;
obj proc;

    save_cont( arg_count_reg, would_be_next );

    literals_reg = continue_intr_tmpl;

    f = OBJ_TO_JUMP_ADDR( gvec_read( literals_reg, SLOT(0) ) );

    /* PUSH_PARTCONT_ADDR() assumes it is at the beginning of a block... */
    {
    PUSH_PARTCONT_ADDR(f,0);
    }

    /* `os_get_next_sig' disables interrupts, too */

    nxt = os_get_next_sig();
    proc = gvec_ref( interrupt_handlers, SLOT(nxt.signum) );
    switch (nxt.signum)
      {
      case RSSIG_USER_INTR:
      case RSSIG_TIMEOUT:
      case RSSIG_GC_FLIP:
	n = 0;
	break;
      case RSSIG_FINALIZE:
	REG0 = nxt.data.finalize.finalize_list;
	n = 1;
	break;
      case RSSIG_C_SIGNAL:
	REG0 = int2fx( nxt.data.c_signal.signal_number );
	n = 1;
	break;
      case RSSIG_CHILD_EXITED:
#if HAVE_WAIT
	{
#ifdef PLATFORM_NEXT

#  define WEXITSTATUS(s) ((s).w_retcode)
#  define WTERMSIG(s) ((s).w_termsig)
	  union wait stat;
	  stat.w_status = nxt.data.child_exited.status;
#else
	  int stat = nxt.data.child_exited.status;
#endif	  
	  REG0 = int2fx( nxt.data.child_exited.process_id );
	  if (WIFEXITED(stat))
	    {
	      REG1 = lookup_symbol( "exited" );
	      REG2 = int2fx( WEXITSTATUS(stat) );
	    }
	  else if (WIFSIGNALED(stat))
	    {
	      REG1 = lookup_symbol( "signalled" );
	      REG2 = int2fx( WTERMSIG(stat) );
	    }
	  n = 3;
	}
#else
      fprintf( stderr, "CHILD_EXITED but didn't wait()" );
      abort();
#endif
      break;
      }
    APPLY( n, proc );
}
