/*-----------------------------------------------------------------*-C-*---
 * File:    handc/runtime/strings.c
 *
 *          Copyright (C)1995 Donovan Kolbly <donovan@tkg.com>
 *          as part of the RScheme project, licensed for free use
 *
 * Version: 1.2
 * Date:    1996.08.30 21:40:46
 * Build:   v0.7-0.9, 97.01.26
 *
 * Purpose: Primitive implementations for basic string operations
 *------------------------------------------------------------------------*/

#include <rscheme/scheme.h>
#include <rscheme/smemory.h>

rs_bool string_eq( obj str1, obj str2 )
{
INT_32 n = SIZEOF_PTR(str1);
const UINT_32 *s1 = (const UINT_32 *)PTR_TO_DATAPTR(str1);
const UINT_32 *s2 = (const UINT_32 *)PTR_TO_DATAPTR(str2);

    if (n != SIZEOF_PTR(str2))
	return NO;

    while (n > 0)
    {
	if (*s1++ != *s2++)
	    return NO;
	n -= 4;
    }
    return YES;
}

rs_bool string_ci_eq( obj str1, obj str2 )
{
    return (string_ci_cmp( str1, str2 ) == 0) ? YES : NO;
}

int string_cmp( obj str1, obj str2 )
{
  UINT_32 len1, len2;
  const unsigned char *s1, *s2;

  s1 = (const unsigned char *)string_text(str1);
  len1 = string_length(str1);

  s2 = (const unsigned char *)string_text(str2);
  len2 = string_length(str2);

  while (1)
    {
      if (len1 && len2)
	{
	  if (*s1 != *s2)
	    {
	      if (*s1 < *s2)
		return -1;
	      else
		return 1;
	    }
	  s1++;
	  s2++;
	  len1--;
	  len2--;
	}
      else
	{
	  if (len1) /* still some str1 left, so str1 is bigger */
	    return 1;
	  else if (len2) /* still some str2 left, so str2 is bigger */
	    return -1;
	  else
	    return 0;
	}
    }
}

int string_ci_cmp( obj str1, obj str2 )
{
  const unsigned char *s1 = (const unsigned char *)string_text(str1);
  const unsigned char *s2 = (const unsigned char *)string_text(str2);
  unsigned char c1, c2;
  UINT_32 len1 = string_length(str1);
  UINT_32 len2 = string_length(str2);

  while (1)
    {
      if (len1 && len2)
	{
	  c1 = *s1++;
	  c2 = *s2++;
	  if (isupper(c1))
	    c1 = tolower(c1);
	  if (isupper(c2))
	    c2 = tolower(c2);
	  if (c1 != c2)
	    {
	      if (c1 < c2)
		return -1;
	      else
		return 1;
	    }
	  len1--;
	  len2--;
	}
      else
	{
	  if (len1)
	    return 1;
	  else if (len2)
	    return -1;
	  else
	    return 0;
	}
    }
}

obj make_string( const char *text )
{
size_t len;
obj str;

    if (text)
    {
	len = strlen(text);
	str = bvec_alloc( len+1, string_class );
	memcpy( PTR_TO_DATAPTR(str), text, len );
	return str;
    }
    else
	return FALSE_OBJ;
}

/* we cast the string text pointer to an 'unsigned char *'
   so that characters we extract do not get sign extended
                                          -- dmk 94.06.21 */

obj string_ref( obj str, UINT_32 index )
{
    assert( index < string_length(str) );
    return MAKE_ASCII_CHAR( ((unsigned char *)string_text(str))[index] );
}

void string_set( obj str, UINT_32 index, obj ch )
{
    assert( index < string_length(str) );
    assert( OBJ_ISA_ASCII_CHAR(ch) );
    ((char *)string_text(str))[index] = GET_IMMEDIATE_VALUE(ch);
}
